#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Fuel segment
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sat Jun 27 14:46:49 EDT 2015
"""
# *******************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random
import pandas

from cortix.support.periodictable import ELEMENTS
from cortix.support.periodictable import SERIES
from cortix.support.specie import Specie
# *******************************************************************************

class FuelSegment():

    # Todo: Species should not be here. Need to replace by Phase instead.
    #      Chopper will be affected

    def __init__(self,
                 geometry = pandas.Series(),
                 species = list()
                 ):

        assert isinstance(geometry, pandas.Series), 'fatal.'
        assert isinstance(species, list), 'fatal.'
        if isinstance(species, list) and len(species) > 0:
            assert isinstance(species[0], Specie)

        self.attribute_names = \
            ['n-segments',
             'fuel-volume',
             'segment-volume',
             'fuel-diameter',
             'fuel-length',
             'mass',
             'mass-dens',
             'mass-cc',
             'nuclides',
             'isotopes',
             'radioactivity',
             'radioactivity-dens',
             'gamma',
             'gamma-dens',
             'heat',
             'heat-dens',
             'molar-heat-pwr',
             'molar-gamma-pwr']

        self.__geometry = geometry
        self.__species  = species
# ---------------------- end def __init__():--------------------------------------

    def get_geometry(self):
        return self.__geometry
    geometry = property(get_geometry, None, None, None)

    def get_species(self):
        return self.__species
    species = property(get_species, None, None, None)

    def get_specie(self, name):
        for specie in self.__species:
            if specie.name == name:
                return specie
        return None
    specie = property(get_specie, None, None, None)

# Get stored fuel segment property either overall or on a nuclide basis
    def get_attribute(self, name, nuclide=None, series=None):

        attribute_name = name

        assert attribute_name in self.attribute_names, \
               'attribute_name: %r; options: %r; fail.' % \
               (attribute_name, self.attribute_names)

        if nuclide is not None:
            assert isinstance(nuclide,str), 'type(nuclide) = %r' % type(nuclide)
            # no multipliers for now (see below: codes is almost ready for it)
            assert len(nuclide.split('*')) == 1, 'nuclide = %r' % nuclide # sanity check

        if nuclide is not None:
            assert series is None, 'fail.'
        if series is not None:
            assert nuclide is None, 'fail.'

        if series is not None:
            assert False, ' not implemented.'

        if attribute_name == 'isotopes':
            assert nuclide is not None, 'need a nuclide symbol.'

# .................................................................................
# # of segments

        if attribute_name == 'n-segments':
            return 1

# .................................................................................
# segment id 

        if attribute_name == 'segment-id':
            return self.__geometry['segment id']

# .................................................................................
# fuel volume

        if attribute_name == 'fuel-volume':
            return self.__get_fuel_segment_volume()
# .................................................................................
# segment volume

        if attribute_name == 'segment-volume':

            cladding_length = self.__geometry['cladding length [cm]']
            cladding_diam = self.__geometry['OD [cm]']
            volume = cladding_length * math.pi * (cladding_diam / 2.0)**2
            return volume

# .................................................................................
# fuel diameter

        if attribute_name == 'fuel-diameter':

            fuel_diam = self.__geometry['fuel diameter [cm]']
            return fuel_diam

# .................................................................................
# fuel length

        if attribute_name == 'fuel-length':

            fuel_length = self.__geometry['fuel length [cm]']
            return fuel_length

# .................................................................................
# fuel segment overall quantities
        if nuclide is None and series is None:

            # mass or mass concentration
            if attribute_name == 'mass-cc' or attribute_name == 'mass-dens' or attribute_name == 'mass':
                mass_cc = 0.0
                for spc in self.__species:
                    mass_cc += spc.massCC
                if attribute_name == 'mass-cc' or attribute_name == 'mass-dens':
                    return mass_cc
                else:
                    volume = self.__get_fuel_segment_volume()
                    return mass_cc * volume
# radioactivity
            if attribute_name == 'radioactivtyDens' or attribute_name == 'radioactivity':
                radDens = 0.0
                for spc in self.__species:
                    radDens += spc.molarRadioactivity * spc.molarCC
                if attribute_name == 'radioactivity-dens':
                    return radDens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return radDens * volume
# gamma
            if attribute_name == 'gamma-dens' or attribute_name == 'gamma':
                gamma_dens = 0.0
                for spc in self.__species:
                    gamma_dens += spc.molarGammaPwr * spc.molarCC
                if attribute_name == 'gamma-dens':
                    return gamma_dens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return gamma_dens * volume
# heat
            if attribute_name == 'heat-dens' or attribute_name == 'heat':
                heat_dens = 0.0
                for spc in self.__species:
                    heat_dens += spc.molarHeatPwr * spc.molarCC
                if attribute_name == 'heat-dens':
                    return heat_dens
                else:
                    volume = self.__get_fuel_segment_volume()
                    return heat_dens * volume

# .................................................................................
# radioactivity

        if attribute_name == 'radioactivity-dens' or attribute_name == 'radioactivity':
            assert False
            colName = 'Radioactivity Dens. [Ci/cc]'

# .................................................................................
# thermal

        if attribute_name == 'thermalDens' or attribute_name == 'thermal' or  \
           attribute_name == 'heat-dens' or attribute_name == 'heat':
            assert False
            colName = 'Thermal Dens. [W/cc]'

# .................................................................................
# gamma

        if attribute_name == 'gamma-dens' or attribute_name == 'gamma':
            assert False
            colName = 'Gamma Dens. [W/cc]'

# .................................................................................
##########################################################################
# .................................................................................

#  if attribute_name[-4:] == 'Dens' or attribute_name[-2:] == 'CC':
#     attributeDens = True
#  else:
#     attributeDens = False

# .................................................................................
# all nuclide content of the fuel added

#  if nuclide is None and series is None:
#
#     density = 0.0
#
#     density = self.propertyDensities[ colName ].sum()
#
#     if attributeDens is False:
#        volume = self.__get_fuel_segment_volume()
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get chemical element series

#  if series is not None:
#
#     density = 0.0
#
#     for isotope in isotopes:
#       density += self.propertyDensities.loc[isotope,colName]
#
#     if attributeDens is False:
#        volume = self.__get_fuel_segment_volume()
#        prop = density * volume
#        return prop
#     else:
#        return density

# .................................................................................
# get specific nuclide (either the isotopes of the nuclide or the specific
# isotope) property: note the most complex case handled is, say: 2*Cs-133.

        if nuclide is not None:

            # a particular nuclide given (atomic number and atomic mass number)
            if len(nuclide.split('-')) == 2:

                nuclideMassNumber = int(nuclide.split('-')[1].strip('m'))
                nuclideSymbol = nuclide.split('-')[0]
                nuclideMolarMass = ELEMENTS[nuclideSymbol].isotopes[nuclideMassNumber].mass

                mass_cc = 0.0

                for spc in self.__species:

                    formula = spc.atoms

                    mole_fraction = 0.0

                    for item in formula:

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            formula_nuclide_symbol = item.split('-')[0].strip()
                            if formula_nuclide_symbol == nuclideSymbol:
                                assert len(item.split('-')) == 2

                            if item.split('*')[0].strip() == nuclide:
                                mole_fraction = 1.0
                            else:
                                mole_fraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            formula_nuclide_symbol = item.split(
                                '*')[1].split('-')[0].strip()
                            if formula_nuclide_symbol == nuclideSymbol:
                                assert len(item.split('*')[1].split('-')) == 2

                            if item.split('*')[1].strip() == nuclide:
                                mole_fraction = float(
                                    item.split('*')[0].strip())
                            else:
                                mole_fraction = 0.0

                        else:
                            assert False

                        mass_cc += spc.molarCC * mole_fraction * nuclideMolarMass

                return mass_cc * self.__get_fuel_segment_volume()

        # chemical element given (only atomic number given)
            elif len(nuclide.split('-')) == 1:

                mass_cc = 0.0

                for spc in self.__species:

                    formula = spc.atoms

                    for item in formula:

                        mole_fraction = 0.0

                        if len(item.split('*')
                               ) == 1:  # no multiplier (implies 1.0)

                            assert len(item.split('-')) == 2
                            formula_nuclide_symbol = item.split('-')[0].strip()
                            formula_nuclide_mass_number = int(
                                item.split('-')[1].strip('m'))
                            formula_nuclide_molar_mass = ELEMENTS[formula_nuclide_symbol].isotopes[formula_nuclide_mass_number].mass

                            if formula_nuclide_symbol == nuclide:
                                mole_fraction = 1.0
                            else:
                                mole_fraction = 0.0

                        elif len(item.split('*')) == 2:  # with multiplier

                            assert len(item.split('*')[1].split('-')) == 2
                            formula_nuclides_symbol = item.split(
                                '*')[1].split('-')[0].strip()
                            formula_nuclide_mass_number = int(
                                item.split('*')[1].split('-')[1].strip('m'))
                            formula_nuclide_molar_mass = ELEMENTS[formula_nuclides_symbol].isotopes[formula_nuclide_mass_number].mass

                            if formula_nuclides_symbol == nuclide:
                                mole_fraction = float(
                                    item.split('*')[0].strip())
                            else:
                                mole_fraction = 0.0

                        else:
                            assert False

                        mass_cc += spc.molarCC * mole_fraction * formula_nuclide_molar_mass

                return mass_cc * self.__get_fuel_segment_volume()

            else:

                assert False

# ---------------------------------------------------------------------------------
    def __get_fuel_segment_volume(self):

        fuel_length = self.__geometry['fuel length [cm]']
        fuel_diam = self.__geometry['fuel diameter [cm]']

        volume = fuel_length * math.pi * (fuel_diam / 2.0)**2

        return volume

# *********************************************************************************

# *******************************************************************************
# Printing of data members
    def __str__(self):
        s = 'FuelSegment(): %s\n %s\n'
        return s % (self.__geometry, self.__species)

    def __repr__(self):
        s = 'FuelSegment(): %s\n %s\n'
        return s % (self.__geometry, self.__species)
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix Module class defintion.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#**********************************************************************************
import os
import logging

from cortix.src.utils.xmltree import XMLTree
from cortix.src.launcher import Launcher
#**********************************************************************************

class Module:
    '''
    The Module class encapsulates a computational module of some scientific domain.
    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__( self, 
            logger,
            importlib_name, library_home_dir, 
            config_xml_tree ):

        assert isinstance(logger, logging.Logger),'ctx::mod: logger is invalid.'
        assert isinstance(importlib_name, str), 'ctx::mod: importlib_name is invalid.'
        assert isinstance(library_home_dir, str),'ctx:mod: library_home is invalid.'

        assert isinstance(config_xml_tree, XMLTree),'ctx:mod: config_xml_tree is invalid.'
        assert config_xml_tree.get_node_tag() == 'module','ctx:mod:invalid module xml tree.'
        # Read the module name and type
        self.__mod_name = config_xml_tree.get_node_attribute('name') # e.g. wind
        self.__mod_type = config_xml_tree.get_node_attribute('type') # e.g. native

        # Specify module library with upstream information
        self.__library_home_dir = library_home_dir
        self.__importlib_name       = importlib_name

        # executable is deprecated; eliminate
        self.__executable_name = 'null-executable_name'
        self.__executable_path = 'null-executable_path'

        self.__input_file_name = 'null-input_file_name'
        self.__input_file_path = 'null-input_file_path'

        # Get config data  
        for child in config_xml_tree.get_node_children():

            (elem, tag, attributes, text) = child
            text = text.strip()

            if self.__mod_type != 'native':
                if tag == 'executable_name':
                    self.__executable_name = text

            if tag == 'executable_path':
                if text[-1] != '/':
                    text += '/'
                self.__executable_path = text

            if tag == 'input_file_name':
                self.__input_file_name = text

            if tag == 'input_file_path':
                if text[-1] != '/':
                    text += '/'
                self.__input_file_path = text

            if tag == 'library':
                assert len(attributes) == 1, 'only name of library allowed.'
                key = attributes[0][0]
                assert key == 'importlib_name', 'invalid attribute key.'
                val = attributes[0][1].strip()
                self.__importlib_name = val

                node = XMLTree( elem ) # fixme: remove this wrapper
                sub_node = node.get_sub_node('home_dir')

                # override home_dir
                # fixme: no root node needed
                self.__library_home_dir = sub_node.get_root_node().text.strip()

                if self.__library_home_dir[-1] == '/':
                    self.__library_home_dir.strip('/')

        logger.debug(self.__mod_name+': read module config info')

        # Take care of a few full path issue
        cortix_path = os.path.abspath(os.path.join(__file__, '../../..'))

        self.__manifest_full_path_file_name = self.__library_home_dir + '/' + \
                self.__mod_name + '/manifest.xml'

        if '$CORTIX' in self.__input_file_path:
            self.__input_file_path = \
                    self.__input_file_path.replace('$CORTIX', cortix_path)

        if '$CORTIX' in self.__manifest_full_path_file_name:
            self.__manifest_full_path_file_name = \
                    self.__manifest_full_path_file_name.replace('$CORTIX', cortix_path)

        # Read the module's manifest
        self.__read_manifest()

        logger.debug(self.__mod_name+': read manifest ports\n %s'%self.__diagram)

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def __get_name(self):
        '''
        `str`:Module name
        '''

        return self.__mod_name

    name = property(__get_name, None, None, None)

    def __get_importlib_name(self):
        '''
        `str`:Module library name
        '''

        return self.__importlib_name

    importlib_name = property(__get_importlib_name, None, None, None)

    def __get_library_home_dir(self):
        '''
        `str`:Library home directory
        '''

        return self.__library_home_dir

    library_home_dir = property(__get_library_home_dir, None, None, None)

    def __get_ports(self):
        '''
       `list(tuple)`: Module's ports
        '''

        return self.__ports

    ports = property(__get_ports, None, None, None)

    def get_port_type(self, port_name):
        '''
        Returns the port type specified by port_name
        '''

        port_type = None
        for port in self.__ports:
            if port[0] == port_name:
                port_type = port[1]

        return port_type

    def get_port_mode(self, port_name):
        '''
        Returns the port mode specified by port_name
        '''

        port_mode = None
        for port in self.__ports:
            if port[0] == port_name:
                port_mode = port[2]

        return port_mode

    def __get_port_names(self):
        '''
        `list(tuple)`:List of names of module's ports
        '''

        port_names = list()
        for port in self.__ports:
            port_names.append(port[0])
        return port_names

    port_names = property(__get_port_names, None, None, None)

    def has_port_name(self, port_name):
        '''
        Returns true if a port with the name
        port_name is available in the module.
        '''

        for port in self.__ports:
            if port[0] == port_name:
                return True

        return False

    def __get_diagram(self):
        '''
        Return the diagram string from the module manifest or a null place holder.
        '''

        return self.__diagram

    diagram = property(__get_diagram, None, None, None)

    def execute( self, slot_id,
                 runtime_cortix_param_file,
                 runtime_cortix_comm_file    ):
        '''
        Spawns a worker process to execute the module.
        '''

        assert runtime_cortix_param_file[-1] is not '/'
        assert runtime_cortix_comm_file[-1] is not '/'

        module_input = self.__input_file_path + self.__input_file_name
        param = runtime_cortix_param_file
        comm  = runtime_cortix_comm_file

        full_path_comm_dir = comm[:comm.rfind('/')] + '/' # extract directory name
        runtime_module_status_file = full_path_comm_dir + 'runtime-status.xml'

        status = runtime_module_status_file

        importlib_name = self.__importlib_name
        library_home_dir = self.__library_home_dir
        mod_name     = self.__mod_name

        # provide a wrk/ for each modules for additional work IO data
        assert os.path.isdir(full_path_comm_dir), \
               'module directory %r not available.' % full_path_comm_dir

        mod_work_dir = full_path_comm_dir + 'wrk/'

        os.system('mkdir -p ' + mod_work_dir)

        assert os.path.isdir(mod_work_dir), \
               'module work directory %r not available.' % mod_work_dir

        # only for wrapped modules (deprecated; remove in the future)
        #mod_exec_name = self.__executable_path + self.__executable_name

        manifest_name = self.__manifest_full_path_file_name

        # the laucher "loads" the module dynamically and provides the method for
        # threading
        launch = Launcher( importlib_name, library_home_dir, mod_name,
                           slot_id,
                           module_input,
                           manifest_name,
                           mod_work_dir,
                           param, comm, status )

        # run module on its own process (file IO communication will take place
        # between modules)
        launch.start() # this start a thread and runs the run() method of launch

        return runtime_module_status_file

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __read_manifest( self ):
        '''
        Get ports
        '''

        assert isinstance(self.__manifest_full_path_file_name, str)

        # Read the manifest 
        xml_tree = XMLTree( xml_tree_file=self.__manifest_full_path_file_name )

        assert xml_tree.get_node_tag() == 'module_manifest'

        assert xml_tree.get_node_attribute('name') == self.__mod_name

        # List of (port_name, port_type, port_mode, port_multiplicity)
        self.__ports = list()

        self.__diagram = 'null-module-diagram'

        # Get config data  
        for child in xml_tree.get_node_children():
            (elem, tag, attributes, text) = child

            if tag == 'port':

                text = text.strip()

                assert len(attributes) == 3, "only <= 3 attributes allowed."

                tmp = dict()  # store port name and three attributes

                for attribute in attributes:
                    key = attribute[0].strip()
                    val = attribute[1].strip()

                    if key == 'type':
                        assert val == 'use' or val == 'provide' or val == 'input' or \
                            val == 'output', 'port attribute value invalid.'
                        tmp['port_name'] = text  # port_name
                        tmp['port_type'] = val   # port_type
                    elif key == 'mode':
                        file_value = val.split('.')[0]
                        assert file_value == 'file' or file_value == 'directory',\
                            'port attribute value invalid.'
                        tmp['port_mode'] = val
                    elif key == 'multiplicity':
                        tmp['port_multiplicity'] = int(val)  # port_multiplicity
                    else:
                        assert False, 'invalid port attribute: %r=%r. fatal.'%\
                                (key,val)

                assert len(tmp) == 4
                store = (tmp['port_name'], tmp['port_type'], tmp['port_mode'],
                         tmp['port_multiplicity'])

                # (port_name, port_type, port_mode, port_multiplicity)
                self.__ports.append(store)

                # clear
                tmp   = None
                store = None

            if tag == 'diagram':

                self.__diagram = text


        return

#======================= end class Module: =======================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Valmor F. de Almeida dealmeidav@ornl.gov; vfda

Pyplot module.

This class manages time-sequence data in XML or tabular formats.
It is a helper for reading and manipulating stored file data in Cortix.
The XML data is a ElementTree object.

Sat Jul 19 12:13:05 EDT 2014
"""
# *********************************************************************************
import os
import sys
import io
import time
import datetime
import logging
import xml.etree.ElementTree as ElementTree
from threading import Lock
# *********************************************************************************

class TimeSequence():

    def __init__(self,
                 fileName,   # full path file name
                 fileType,   # "xml"
                 initialTime=0.0,
                 finalTime=0.0,
                 time_unit=None,
                 logger=None
                 ):

        assert isinstance(fileName, str), 'wrong type; stop.'
        self.__fileName = fileName

        assert isinstance(fileType, str), 'wrong type; stop.'
        self.__fileType = fileType

        assert isinstance(initialTime, float), 'wrong type; stop.'
        self.__initialTime = initialTime

        assert isinstance(finalTime, float), 'wrong type; stop.'
        self.__finalTime = finalTime

        assert isinstance(time_unit, str), 'wrong type; stop.'
        self.__time_unit = time_unit

        assert finalTime >= initialTime, 'sanity check failed. stop.'
        assert initialTime >= 0.0, 'sanity check failed. stop.'
        assert finalTime >= 0.0, 'sanity check failed. stop.'

        assert isinstance(logger, logging.Logger), 'wrong type; stop.'
        assert logger is not None, 'must give a logger; stop.'
        self.log = logger

        self.__tree = None

        if fileType == 'xml':
            self.__read_xml()

#  s = 'TimeSequence::__init__(): built object'
#  self.log.debug(s)
# ---------------------- end def __init__():------------------------------

# ---------------------------------------------------------------------------------
# Accessors (note: all accessors of member data can potentially change the
# python object referenced to). See NB below.

    def get_name(self):
        root = self.__tree.getroot()
        return root.get('name').strip()

    def GetTimeUnit(self):
        root = self.__tree.getroot()
        timeNode = root.find('time')
        return timeNode.get('unit').strip()

    def GetNVariables(self):
        root = self.__tree.getroot()
        varNodes = root.findall('var')
        return len(varNodes)

    def GetVariableNames(self):
        names = list()
        root = self.__tree.getroot()
        varNodes = root.findall('var')
        for v in varNodes:
            name = v.get('name').strip()
            names.append(name)
        return names

    def GetVariables(self):
        # variables[(name,unit,timeUnit,legend)] = [(time,val),(time,val),..,]
        variables = dict()
        root = self.__tree.getroot()
        timeNode = root.find('time')
        timeUnit = timeNode.get('unit').strip()
        varNodes = root.findall('var')
        timeStampNodes = root.findall('timeStamp')
        for ivar in range(len(varNodes)):
            name = varNodes[ivar].get('name').strip()
            unit = varNodes[ivar].get('unit').strip()
            legend = varNodes[ivar].get('legend').strip()
            scale = varNodes[ivar].get('scale')
            if scale is None:
                spec = (name, unit, timeUnit, legend, 'linear')
            else:
                scale = scale.strip()
                assert scale == 'log' or scale == 'linear' or scale == 'log-log' or \
                    scale == 'linear-linear' or scale == 'log-linear' or \
                    scale == 'linear-log'
                spec = (name, unit, timeUnit, legend, scale)
            timeValues = list()
            for ts in timeStampNodes:
                time = float(ts.get('value').strip())
                if time >= self.__initialTime and time <= self.__finalTime:
                    data = ts.text.strip().split(',')
                    assert len(
                        data) >= 1, 'empty data field in the %r variable; stop.' % name
# Accept missing data and fill in as zero; or neglect excess of data
#        if ivar == 0:
# assert len(data) == len(varNodes), '# variables %r != # values %r' %
# (len(data),len(varNodes))
                    if len(data) >= len(varNodes):
                        timeValues.append((time, float(data[ivar])))
                    else:
                        timeValues.append((time, float(0.0)))
            variables[spec] = timeValues

        return variables

# *********************************************************************************
# Private helper functions (internal use: __)

    def __read_xml(self):
        '''
        Read the xml data file recursively until the desired time stamp is found.
        '''

        s = 'TimeSequence::__read_xml(): try reading: ' + self.__fileName
        self.log.debug(s)

        found = False

        while found is False:

            try:
                mutex = Lock()
                mutex.acquire()
                tree = ElementTree.parse(self.__fileName)
            except ElementTree.ParseError as error:
                mutex.release()
                s = 'TimeSequence(): ' + self.__fileName + ' unavailable. Error code: ' + \
                    str(error.code) + '; File position: ' + \
                    str(error.position) + '. Retrying...'
                self.log.debug(s)
                time.sleep(0.1)
                continue

            mutex.release()
            self.__tree = tree
            rootNode = self.__tree.getroot()
            assert rootNode.tag == 'time-sequence', 'invalid format.'

            node = rootNode.find('time')
            timeUnit = node.get('unit').strip()

            time_unit_scale = 1.0
#            time_unit_scale = 60.0 # __get_time_unit_scale( timeUnit )

            # (cut-off) legacy stuff
            timeCutOff = node.get('cut-off')
            if timeCutOff is not None:
                timeCutOff = float(timeCutOff.strip())
                if self.__finalTime > timeCutOff:
                    return

            nodes = rootNode.findall('timeStamp')

            for n in nodes:

                timeStamp = float(n.get('value').strip())

                if abs(timeStamp - self.__finalTime) <= 1.0e-5:
                    return

        # end of while found is False

        return
# ---------------------- end def __read_xml():----------------------------



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Simulation class of Cortix.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import logging
import datetime
from cortix.src.utils.xmltree       import XMLTree
from cortix.src.utils.set_logger_level import set_logger_level

from cortix.src.task        import Task
from cortix.src.application import Application
#*********************************************************************************

class Simulation:
    '''
    Cortix Simulation element as defined in the Cortix config.
    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__(self, parent_work_dir, config_xml_tree):

        assert isinstance(parent_work_dir, str), '-> parent_work_dir invalid.'

        assert isinstance(config_xml_tree, XMLTree), '-> config_xml_tree invalid.'
        assert config_xml_tree.get_node_tag() == 'simulation'

        # Read the simulation name, e.g. <simulation name='droplet'></simulation>
        self.__name = config_xml_tree.get_node_attribute('name')

        # Create the cortix/simulation work directory
        self.__work_dir = parent_work_dir + 'sim_' + self.__name + '/'

        os.system('mkdir -p ' + self.__work_dir)

        # Create the logging facility 
        self.__create_logging( config_xml_tree )

        #======================
        # Create application(s)
        #======================
        for app_config_xml_node in config_xml_tree.get_all_sub_nodes('application'):

            assert app_config_xml_node.get_node_tag() == 'application'

            # Create the application for this simulation (same work directory)
            self.__application = Application( self.__work_dir, app_config_xml_node )

            self.__log.debug("created application: %s",
                    app_config_xml_node.get_node_attribute('name'))

        # Stores the task(s) created by the execute method
        self.__tasks = list() # not needed now; in the future for task parallelism

        # Save configuration information for setting up task in execute()
        self.__config_xml = config_xml_tree

        self.__log.info("created simulation: %s", self.__name)

        return

    def __del__(self):

        self.__log.info("destroyed simulation: %s", self.__name)

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def execute(self, task_name=None):
        '''
        This method allows for the execution of a simulation by executing each
        task, if any. Execution proceeds one task at a time.
        '''

        self.__log.debug("prepare to execute(%s)", task_name)

        if task_name is not None:

            # Create the task object for each task
            self.__setup_task( task_name )

            for task in self.__tasks:

                if task.name == task_name:

                    self.__log.debug("call execute(%s)", task_name)

                    task.execute(self.__application)

                    self.__log.debug(
                        'called task.execute() on task %s', task_name)

        self.__log.debug("end execute(%s)", task_name)

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __create_logging( self, config_xml_tree ):
        '''
        A helper function to setup the logging facility used in self.__init__()
        '''

        logger_name = 'sim:' + self.__name

        self.__log = logging.getLogger(logger_name)
        self.__log.setLevel(logging.NOTSET)

        node = config_xml_tree.get_sub_node("logger")

        logger_level = node.get_node_attribute('level')
        self.__log = set_logger_level(self.__log, logger_name, logger_level)

        file_handler = logging.FileHandler(self.__work_dir + 'sim.log')
        file_handler.setLevel(logging.NOTSET)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.NOTSET)

        for child in node.get_node_children():
            (elem,tag,attributes,text) = child
            elem = XMLTree( elem ) # fixme: remove wrapping
            if tag == 'file_handler':
                file_handler_level = elem.get_node_attribute('level')
                file_handler = set_logger_level(file_handler, logger_name,
                                                file_handler_level)

            if tag == 'console_handler':
                console_handler_level = elem.get_node_attribute('level')
                console_handler = set_logger_level(console_handler, logger_name,
                                                   console_handler_level)

        # formatter added to handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        # add handlers to logger
        self.__log.addHandler(file_handler)
        self.__log.addHandler(console_handler)

        self.__log.info('created Simulation logger: %s', self.__name)
        self.__log.debug('logger level: %s', logger_level)
        self.__log.debug('logger file handler level: %s', file_handler_level)
        self.__log.debug('logger console handler level: %s', console_handler_level)

        return

    def __setup_task(self, task_name):
        '''
        This is a helper function used by the execute() method. It createas a
        Task object and sets up the communication file for the task at hand.
        There must be only one nework for this task and this network will be
        used to generate the communication file for each module where the ports
        will point to files. Each module will have in its runtime directory an
        XML file which maps a port name and type to a file. If the port type is a
        provide type, the file pointed to resides on the same directory as the
        communication file. If the port type is a use type, the file pointed to
        resides in the directory of the module whose corresponding provide port is
        connected to. The communication XML file contains a root tag
        <cortix_comm>, and a series of XML port tags with three attributes and no
        content as follows:

                <port name=port_name type=port_type file=full_path_file_name />
        '''

        self.__log.debug('start __setup_task(%s)', task_name)

        task = None  # flag

        # loop over all elements with a <task></task> tag in this simulation 
        # __config_xml
        for task_config_xml_node in self.__config_xml.get_all_sub_nodes('task'):

            if task_config_xml_node.get_node_attribute('name') != task_name:
                continue

            # create task
            task = Task( self.__work_dir, task_config_xml_node )

            self.__tasks.append(task)

            self.__log.debug('appended task: %s', task_config_xml_node.get_node_attribute('name'))

        if task is None:
            self.__log.info('no task to exectute; done here.')
            self.__log.info('end __setup_task(%s)', task_name)
            return

        # create subdirectory with task name
        task_name = task.name
        task_work_dir = task.work_dir
        assert os.path.isdir(task_work_dir), 'directory %r invalid.' % task_work_dir

        # set the parameters for the task in the cortix param file
        task_file = task_work_dir + 'cortix-param.xml'

        fout = open(task_file, 'w')
        fout.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        fout.write('<!-- Written by Simulation::__setup_task() -->\n')
        today = datetime.datetime.today()
        fout.write('<!-- ' + str(today) + ' -->\n')

        fout.write('<cortix_param>\n')

        start_time = task.start_time
        start_time_unit = task.start_time_unit
        fout.write('<start_time unit="' + start_time_unit + '"' + '>' +
                           str(start_time) + '</start_time>\n')

        evolve_time = task.evolve_time
        evolve_time_unit = task.evolve_time_unit
        fout.write('<evolve_time unit="' + evolve_time_unit + '"' + '>' +
                           str(evolve_time) + '</evolve_time>\n')

        time_step = task.time_step
        time_step_unit = task.time_step_unit
        fout.write('<time_step unit="' + time_step_unit + '"' + '>' +
                           str(time_step) + '</time_step>\n')

        fout.write('</cortix_param>')
        fout.close()

        task.set_runtime_cortix_param_file(task_file)

        # Using the task and network objects create the runtime module directories and 
        # cortix comm files; one comm file per module inside the module runtime directory

        networks = self.__application.networks

        for net in networks:

            if net.name != task_name:
                s = '__setup_task():: net name: ' + \
                    net.name + ' not equal to task name: ' + task_name + '; ignored.'
                self.__log.warn(s)
                continue

            elif net.name == task_name:  # net and task names must match

                connect = net.connectivity
                to_module_to_port_visited = dict()  # providers

                for con in connect:
                    # Start with the ports that will function as a provide port or
                    # input port
                    to_module_slot = con['toModuleSlot']
                    to_port = con['toPort']

                    if to_module_slot not in to_module_to_port_visited.keys():
                        to_module_to_port_visited[to_module_slot] = list()

                    to_module_name = '_'.join( to_module_slot.split('_')[:-1] )
                    to_module = self.__application.get_module(to_module_name)

                    assert to_module is not None, \
                        'module %r does not exist in application' % to_module_name

                    assert to_module.has_port_name(to_port), \
                        'module %r has no port %r.' % (
                            to_module.name, to_port)

                    assert to_module.get_port_type(to_port) is not None,\
                        'network name: %r, module name: %r, to_port: %r port type \
                    invalid %r' % (net.name, to_module.name, to_port,
                                   type(to_module.get_port_type(to_port)))

                    to_module_slot_work_dir = task_work_dir + to_module_slot + '/'

                    if to_module.get_port_type(to_port) != 'input':
                        assert to_module.get_port_type(to_port) == 'provide', \
                            'port type %r invalid. Module %r, port %r' \
                            % (to_module.get_port_type(to_port), to_module.name,
                               to_port)

                    # "to" is who receives the "call", hence the provider
                    to_module_slot_comm_file = to_module_slot_work_dir + \
                        'cortix-comm.xml'

                    if not os.path.isdir(to_module_slot_work_dir):
                        os.system('mkdir -p ' + to_module_slot_work_dir)

                    if not os.path.isfile(to_module_slot_comm_file):
                        fout = open(to_module_slot_comm_file,'w')
                        fout.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                        fout.write('<!-- Written by Simulation::__setup_task() -->\n')
                        today = datetime.datetime.today()
                        fout.write('<!-- ' + str(today) + ' -->\n')
                        fout.write('<cortix_comm>\n')

                    if to_port not in to_module_to_port_visited[to_module_slot]:
                        fout = open(to_module_slot_comm_file,'a')
                        # this is the cortix info for modules providing data
                        to_port_mode = to_module.get_port_mode(to_port)
                        if to_port_mode.split('.')[0] == 'file':
                            ext = to_port_mode.split('.')[1]
                            log_str = '<port name="' + to_port + \
                                    '" type="provide" file="' + to_module_slot_work_dir\
                                    + to_port + '.' + ext + '"/>\n'
                        elif to_port_mode == 'directory':
                            log_str = '<port name="' + to_port + \
                                    '" type="provide" directory="' + \
                                    to_module_slot_work_dir + to_port + '"/>\n'
                        else:
                            assert False, 'invalid port mode. fatal.'

                        fout.write(log_str)
                        fout.close()
                        to_module_to_port_visited[to_module_slot].append(
                            to_port)

                    debug_str = '__setup_task():: comm module: ' + to_module_slot + \
                                '; network: ' + task_name + ' ' + log_str
                    self.__log.debug(debug_str)

                    # register the cortix-comm file for the network
                    net.set_runtime_cortix_comm_file_name( to_module_slot,
                        to_module_slot_comm_file)

                    # Now do the ports that will function as use ports or output ports
                    from_module_slot = con['fromModuleSlot']
                    from_port = con['fromPort']
                    from_module_name = '_'.join( from_module_slot.split('_')[:-1] )
                    from_module = self.__application.get_module(from_module_name)

                    assert from_module.has_port_name(from_port), \
                        'module %r has no port %r'%(from_module_name, from_port)

                    assert from_module.get_port_type(from_port) is not None, \
                        'network name: %r, module name: %r, from_port: %r port type invalid %r'\
                        % (net.name, from_module.name, from_port,
                           type(from_module.get_port_type(from_port)))

                    from_module_slot_work_dir = task_work_dir + from_module_slot + '/'

                    if from_module.get_port_type(from_port) != 'output':
                        assert from_module.get_port_type(from_port) == 'use', 'port type %r invalid. Module %r, port %r' % (from_module.get_port_type(from_port), from_module.name, from_port)

                        # "from" is who makes the "call", hence the user
                        from_module_slot_comm_file = from_module_slot_work_dir + \
                                'cortix-comm.xml'

                        if not os.path.isdir(from_module_slot_work_dir):
                            os.system('mkdir -p ' + from_module_slot_work_dir)

                        if not os.path.isfile(from_module_slot_comm_file):
                            fout = open(from_module_slot_comm_file,'w')
                            fout.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                            fout.write('<!-- Written by Simulation::__setup_task() -->\n')
                            today = datetime.datetime.today()
                            fout.write('<!-- ' + str(today) + ' -->\n')
                            fout.write('<cortix_comm>\n')
                            fout.close() # is this right?

                        fout = open(from_module_slot_comm_file,'a')

                        # this is the cortix info for modules using data
                        assert from_module.get_port_type(from_port) == 'use', \
                            'from_port must be use type.'

                        to_port_mode = to_module.get_port_mode(to_port)
                        if to_port_mode.split('.')[0] == 'file':
                            ext = to_port_mode.split('.')[1]
                            log_str = '<port name="' + from_port + \
                                      '" type="use" file="' + \
                                      to_module_slot_work_dir + to_port + '.' + ext + \
                                      '"/>\n'
                        elif to_port_mode == 'directory':
                            log_str = '<port name="' + from_port + \
                                      '" type="use" directory="' + \
                                      to_module_slot_work_dir + to_port + '"/>\n'
                        else:
                            assert False, 'invalid port mode. fatal.'

                        fout.write(log_str)
                        fout.close()

                        debug_str = '__setup_task():: comm module: ' + from_module_slot\
                                + '; network: ' + task_name + ' ' + log_str
                        self.__log.debug(debug_str)

                # register the cortix-comm file for the network
                net.set_runtime_cortix_comm_file_name( from_module_slot,
                                                  from_module_slot_comm_file )

            #end elif net.name.strip() == task_name.strip():  # net and task names must match
            else:
                assert False, 'this should not happen...'

        # end for net in networks:

        # Now finish forming the XML comm file for each module in each network
        for net in networks:
            for slot_name in net.module_slot_names:
                comm_file = net.get_runtime_cortix_comm_file_name(slot_name)
                if comm_file == 'null-runtime_cortix_comm_file_name':
                    if net.name != task_name:
                       s = '__setup_task():: comm file for ' + slot_name + \
                           ' in network ' + net.name + ' is ' + comm_file
                       self.__log.warn(s)
                       continue
                    else:
                       assert False, 'FATAL ERROR building the comm file for module %r'%\
                                    slot_name
                fout = open(comm_file, 'a')
                fout.write('</cortix_comm>')
                fout.close()

        self.__log.debug('end __setup_task(%s)', task_name)

        return

#======================= end class Simulation: ===================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
This FuelBundle class is a container for usage with other plant-level process modules.
It is meant to represent a fuel bundle of an oxide fuel LWR reactor.
There are three main data structures:

  1) fuel bundle specs
  2) solid phase
  3) gas phase

The container user will have to provide all the data and from then on, this class
will help acess the data.
The printing methods reveal the contained data.

Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sun Dec 27 15:06:55 EST 2015
"""
# *******************************************************************************
import os
import sys
import math
import pandas
from copy import deepcopy
# *******************************************************************************

class FuelBundle():

    def __init__(self,
                 specs=pandas.DataFrame()
                 ):

        assert isinstance(specs, pandas.DataFrame), 'oops not pandas table.'

        self.__specs = specs

        self.__solid_phase = None
        self.__gas_phase = None

        return
# ---------------------- end def __init__():------------------------------

    # ------
    # Start: Pre-irradiation information

    def get_name(self):
        return self.__specs.loc['Name', 1]
    name = property(get_name, None, None, None)

    def get_fuel_enrichment(self):
        return self.__get_fuel_enrichment()
    fuel_enrichment = property(get_fuel_enrichment, None, None, None)

    def get_fresh_u_mass(self):
        return self.__get_fresh_u_mass()
    fresh_u_mass = property(get_fresh_u_mass, None, None, None)

    def get_fresh_u238_mass(self):
        total_mass = self.__get_fresh_u_mass()
        fuel_enrichment = self.__get_fuel_enrichment()
        return total_mass * (1.0 - fuel_enrichment / 100.0)
    fresh_u238_mass = property(get_fresh_u238_mass, None, None, None)

    def get_fresh_U235_mass(self):
        total_u_Mass = self.__get_fresh_u_mass()
        fuel_enrichment = self.__get_fuel_enrichment()
        return total_u_mass * fuel_enrichment / 100.0
    fresh_u235_mass = property(get_fresh_U235_mass, None, None, None)

    def get_n_fuel_rods(self):
        return self.__get_n_fuel_rods()
    n_fuel_rods = property(get_n_fuel_rods, None, None, None)

    # End: Pre-irradiation information
    # ------

    def get_fuel_pin_length(self):
        return self.__get_fuel_pin_length()

    def set_fuel_pin_length(self, x):
        self.__specs.loc['Fuel rods fuel length [in]', 1] = x / 2.54  # in
        return
    fuel_pin_length = property(get_fuel_pin_length, set_fuel_pin_length, None, None)

    def get_fuel_rod_od(self):
        return self.__get_fuel_rod_od()
    fuel_rod_od = property(get_fuel_rod_od, None, None, None)

    def get_fuel_pin_radius(self):
        return self.__get_fuel_pin_radius()
    fuel_pin_radius = property(get_fuel_pin_radius, None, None, None)

    def get_fuel_pin_volume(self):
        return self.__get_fuel_pin_volume()
    fuel_pin_volume = property(get_fuel_pin_volume, None, None, None)

    def get_fuel_volume(self):
        return self.__get_fuel_volume()
    fuel_volume = property(get_fuel_volume, None, None, None)

    # mass of the solid phase (gas phase in plenum not added)
    def get_fuel_mass(self):
        return self.__solid_phase.GetValue('mass')
    fuel_mass = property(get_fuel_mass, None, None, None)

    def get_fuel_mass_unit(self):
        return self.__solid_phase.GetQuantity('mass').unit
    fuel_mass_unit = property(get_fuel_mass_unit, None, None, None)

    def get_gas_mass(self):
        return self.__gas_phase.GetValue('mass')
    gas_mass = property(get_gas_mass, None, None, None)

    def get_radioactivity(self):
        return self.__solid_phase.GetValue('radioactivity') + \
            self.__gas_phase.GetValue('radioactivity')
    radioactivity = property(get_radioactivity, None, None, None)

    def get_gamma_pwr(self):
        return self.__solid_phase.GetValue('gamma') + \
            self.__gas_phase.GetValue('gamma')
    gamma_pwr = property(get_gamma_pwr, None, None, None)

    def get_heat_pwr(self):
        return self.__solid_phase.GetValue('heat') + \
            self.__gas_phase.GetValue('heat')
    heat_pwr = property(get_heat_pwr, None, None, None)

    def get_fuel_radioactivity(self):       # radioactivity of the solid phase
        return self.__solid_phase.GetValue('radioactivity')
    fuel_radioactivity = property(get_fuel_radioactivity, None, None, None)

    def get_gas_radioactivity(self):  # radioactivity of the gas phase
        return self.__gas_phase.GetValue('radioactivity')
    gas_radioactivity = property(get_gas_radioactivity, None, None, None)

    def get_solid_phase(self):
        return self.__solid_phase

    def set_solid_phase(self, phase):
        self.__solid_phase = deepcopy(phase)
    solid_phase = property(get_solid_phase, set_solid_phase, None, None)

    def get_gas_phase(self):
        return self.__gas_phase

    def set_gas_phase(self, phase):
        self.__gas_phase = deepcopy(phase)
    gas_phase = property(get_gas_phase, set_gas_phase, None, None)

# *********************************************************************************
# Private helper functions (internal use: __)

    def __get_fuel_enrichment(self):
        return float(self.__specs.loc['Enrichment [U-235 wt%]', 1])

    def __get_fresh_u_mass(self):
        return float(
            self.__specs.loc['U mass per assy [kg]', 1]) * 1000.0  # [g]

    def __get_n_fuel_rods(self):
        return int(self.__specs.loc['Fuel rods number', 1])

    def __get_fuel_pin_length(self):
        return float(
            self.__specs.loc['Fuel rods fuel length [in]', 1]) * 2.54  # cm

    def __get_fuel_rod_od(self):
        return float(self.__specs.loc['Fuel rods O.D. [in]', 1]) * 2.54  # cm

    def __get_fuel_rod_wall_thickness(self):
        return float(
            self.__specs.loc['Fuel rods wall thickness [in]', 1]) * 2.54  # cm

    def __get_fuel_pin_radius(self):
        fuel_rod_od = self.__get_fuel_rod_od()
        fuel_rod_wall_thickness = self.__get_fuel_rod_wall_thickness()
        fuel_pin_radius = (fuel_rod_od - 2.0 * fuel_rod_wall_thickness) / 2.0
        return fuel_pin_radius

    def __get_fuel_pin_volume(self):
        fuel_pin_length = self.__get_fuel_pin_length()
        fuel_pin_radius = self.__get_fuel_pin_radius()
        return fuel_pin_length * math.pi * fuel_pin_radius ** 2

    def __get_fuel_volume(self):
        fuel_pin_volume = self.__get_fuel_pin_volume()
        n_fuel_rods = self.__get_n_fuel_rods()
        return fuel_pin_volume * n_fuel_rods

# ********************************************************************************
# Printing of data members
    def __str__(self):
        s = 'FuelBundle():\n %s\n %s\n %s\n'
        return s % (self.__specs, self.__solid_phase, self.__gas_phase)

    def __repr__(self):
        s = 'FuelBundle():\n %s\n %s\n %s\n'
        return s % (self.__specs, self.__solid_phase, self.__gas_phase)

# ====================== end class FuelBundle: ===================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda

Stream container

VFdALib support classes

Sat Aug 15 17:24:02 EDT 2015
"""

# *******************************************************************************
import os
import sys
import pandas

from cortix.support.specie import Specie
from cortix.support.quantity import Quantity
# *******************************************************************************

# *******************************************************************************


class Stream():

    def __init__(self,
                 timeStamp,
                 species=None,
                 quantities=None,
                 values=float(0.0)
                 ):

        # constructor
        # THIS WILL CREATE AN ORDERED "STREAM"; values must be in order!

        assert isinstance(timeStamp, float)

        self.timeStamp = timeStamp

        if species is not None:
            assert isinstance(species, list)
            if len(species) > 0:
                assert isinstance(species[-1], Specie)

        if quantities is not None:
            assert isinstance(quantities, list)
            if len(quantities) > 0:
                assert isinstance(quantities[-1], Quantity)

#  assert type(value) == type(float())

# List of quantities and species objects
        self.species = species       # list
        self.quantities = quantities    # list

        names = list()  # note order of names; values must be in the same order; caution

        for specie in self.species:
            names.append(specie.name)

        for quant in self.quantities:
            names.append(quant.name)

# ORDERED data; caution!!
# Table data stream
        self.stream = pandas.DataFrame(index=[timeStamp], columns=names)
        if isinstance(values, float):
            self.stream.fillna(values, inplace=True)
        else:
            self.stream.fillna(0.0, inplace=True)

        if isinstance(values, list) and len(values) == len(names):
            for (name, val) in zip(names, values):
                self.stream.loc[timeStamp, name] = val

        return


# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def GetTimeStamp(self):
        return self.timeStamp

    def GetActors(self):
        return list(self.stream.columns)

    def GetSpecie(self, name):
        for specie in self.species:
            if specie.name == name:
                return specie
        return None

    def GetSpecies(self):
        return self.species

    def GetQuantities(self):
        return self.quantities

    def SetSpecieId(self, name, val):
        for specie in self.species:
            if specie.name == name:
                specie.flag = val
                return

    def GetQuantity(self, name):
        for quant in self.quantities:
            if quant.name == name:
                return quant
        return None

    def GetRow(self, timeStamp=None):
        if timeStamp is None:
            return list(self.stream.loc[self.timeStamp, :])
        else:
            assert timeStamp in self.stream.index, 'timeStamp = %r' % (
                timeStamp)
            return list(self.stream.loc[timeStamp, :])

    def GetValue(self, actor, timeStamp=None):
        assert actor in self.stream.columns
        if timeStamp is None:
            return self.stream.loc[self.timeStamp, actor]
        else:
            assert timeStamp in self.stream.index
            return self.stream.loc[timeStamp, actor]

    def SetValue(self, actor, value=None, timeStamp=None):
        assert actor in self.stream.columns
        if timeStamp is None:
            if value is None:
                self.stream.loc[self.timeStamp, actor] = float(0.0)
            else:
                self.stream.loc[self.timeStamp, actor] = float(value)
        else:
            assert timeStamp in self.stream.index, 'timeStamp = %r' % (
                timeStamp)
            if value is None:
                self.stream.loc[timeStamp, actor] = float(0.0)
            else:
                self.stream.loc[timeStamp, actor] = float(value)

# *******************************************************************************
# Printing of data members
# def __str__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
#
# def __repr__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Simple MyModule module template for developers.
'''
#*********************************************************************************
import os, sys, io, time
import logging
#*********************************************************************************

class MyModule():
 r'''
  MyModule template for Cortix.
 '''

 def __init__( self,
               slot_id,
               input_full_path_file_name,
               work_dir,
               ports = list(),
               cortix_start_time = 0.0,
               cortix_final_time = 0.0,
               cortix_time_unit  = None 
             ):

#.................................................................................
# Sanity test

  assert isinstance(slot_id, int), '-> slot_id type %r is invalid.' % type(slot_id)
  assert isinstance(ports, list), '-> ports type %r is invalid.'  % type(ports)
  assert len(ports) > 0
  assert isinstance(cortix_start_time,float), '-> time type %r is invalid.' % \
         type(cortix_start_time)
  assert isinstance(cortix_final_time, float), '-> time type %r is invalid.' % \
         type(cortix_final_time)
  assert isinstance(cortix_time_unit, str), '-> time type %r is invalid.' % \
         type(cortix_time_unit)

  # Logging
  self.__log = logging.getLogger('launcher-mymodule_'+str(slot_id)+'.cortix_driver.mymodule')
  self.__log.info('initializing an object of MyModule()')

#.................................................................................
# Member data 

  self.__slot_id = slot_id
  self.__ports  = ports

  if work_dir[-1] != '/': work_dir = work_dir + '/'
  self.__wrkDir = work_dir

  # Signal to start operation
  self.__goSignal = True    # start operation immediately
  for port in self.__ports: # if there is a signal port, start operation accordingly
      (portName,portType,thisPortFile) = port
      if portName == 'go-signal' and portType == 'use': self.__goSignal = False

  self.__setup_time = 1.0  # min; a delay time before starting to run

#.................................................................................
# Input ports
# Read input information if any

  #fin = open(input_full_path_file_name,'r')

#---------------------- end def __init__():---------------------------------------

 def call_ports( self, cortix_time=0.0 ):
  '''
  Developer must implement this method.
  Transfer data at cortix_time
  '''

  # provide data using the 'provide-port-name' of the module
  #self.__provide_data( provide_port_name='provide-port-name', at_time=cortix_time )

  # use data using the 'use-port-name' of the module
  #self.__use_data( use_port_name='use-port-name', at_time=cortix_time )

  return
#---------------------- end def call_ports():---------------------------------------

 def execute( self, cortix_time=0.0, cortix_time_step=0.0 ):
  '''
  Developer must implement this method.
  Evolve system from cortix_time to cortix_time + cortix_time_step
  '''

  s = 'execute('+str(round(cortix_time,2))+'[min]): '
  self.__log.debug(s)

  # Developer implements helper method, for example
  #self.__evolve( self, cortix_time, cortix_time_step ):

  return
#---------------------- end def execute():----------------------------------------

#*********************************************************************************
# Private helper functions (internal use: __)

 def __provide_data( self, provide_port_name=None, at_time=0.0 ):
  '''
  Example of how this internal method would look like
  '''

# Access the port file
  port_file = self.__get_port_file( provide_port_name = provide_port_name )

# Provide data to port files
  #if provide_port_name == 'provide-port-name' and port_file is not None: 
  #   self.__provide_mymodule_method( port_file, at_time )

  return
#---------------------- end def __provide_data():---------------------------------

 def __use_data( self, use_port_name=None, at_time=0.0 ):
  '''
  Example of how this internal method would look like
  '''

# Access the port file
  port_file = self.__get_port_file( use_port_name = use_port_name )
 
# Use data from port file
  #if use_port_name == 'use-port-name' and port_file is not None:  
  #   self.__use_mymodule_method( port_file, at_time )

  return
#---------------------- end def __use_data():-------------------------------------

 def __get_port_file( self, use_port_name=None, provide_port_name=None ):
   """
   This may return a None port_file
   """

   port_file = None

   #..........
   # Use ports
   #..........
   if use_port_name is not None:
 
     assert provide_port_name is None
 
     for port in self.__ports:
       (portName,portType,thisPortFile) = port
       if portName == use_port_name and portType == 'use': port_file = thisPortFile

     if port_file is None: return None
 
     max_n_trials = 50
     n_trials     = 0
     while os.path.isfile(port_file) is False and n_trials <= max_n_trials:
       n_trials += 1
       time.sleep(0.1)
 
     if n_trials > max_n_trials:
         s = '__get_port_file(): waited ' + str(n_trials) + ' trials for port: ' +\
             port_file
         self.__log.warn(s)
 
     assert os.path.isfile(port_file) is True, \
            'port_file %r not available; stop.' % port_file

   #..............
   # Provide ports
   #..............
   if provide_port_name is not None:
 
     assert use_port_name is None

     for port in self.__ports:
       (portName,portType,thisPortFile) = port
       if portName == provide_port_name and portType == 'provide': port_file = thisPortFile
 
   return port_file
#---------------------- end def __get_port_file():--------------------------------

#======================= end class MyModule: =====================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
This FuelBucket class is a container for usage with other plant-level process modules.
It is meant to represent a fuel bucket of a metal fuel reactor.
----------
ATTENTION:
----------
This container uses Phase() for phases (cladding and fuel). Therefore user is
responsible to make the "history" of the phases consistent. See Phase() info.

Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
"""
#*************************************************************************
import os
import sys
import math
import pandas
from copy import deepcopy
#*************************************************************************

class FuelBucket():

    def __init__(self,
                 specs=pandas.DataFrame()
                ):

        assert isinstance(specs, pandas.DataFrame), 'oops not pandas table.'

        self.__specs = specs

        self.__solid_phase = None

        self.__fuel_phase = None
        self.__cladding_phase = None

        return
 # ---------------------- end def __init__():------------------------------

    #------
    # Start: Pre-irradiation information

    def get_name(self):
        return self.__specs.loc['Name', 1]
    name = property(get_name, None, None, None)

    def get_slug_type(self):
        return self.__specs.loc['Slug type', 1]
    slug_type = property(get_slug_type, None, None, None)

    def get_n_slugs(self):
        return self.__get_n_slugs()
    n_slugs = property(get_n_slugs, None, None, None)

    def get_fuel_enrichment(self):
        return self.__get_fuel_enrichment()
    fuel_enrichment = property(get_fuel_enrichment, None, None, None)

    def get_fresh_u_mass(self):
        return self.__get_fresh_u_mass()
    fresh_u_mass = property(get_fresh_u_mass, None, None, None)

    def get_fresh_u238_mass(self):
        return self.__get_fresh_u238_mass()
    fresh_u238_mass = property(get_fresh_u238_mass, None, None, None)

    def get_fresh_u235_mass(self):
        return self.__GetFreshU235UMass()
    fresh_u235_mass = property(get_fresh_u235_mass, None, None, None)

    def get_cladding_mass(self):
        return self.__get_cladding_mass()
    cladding_mass = property(get_cladding_mass, None, None, None)

    # End: Pre-irradiation information
    #------

    def get_slug_length(self):
        return self.__get_slug_length()

    def set_slug_length(self, x):
        self.__set_slug_length(x)
    slug_length = property(get_slug_length, set_slug_length, None, None)

    def get_outer_slug_od(self):
        return self.__get_outer_slug_od()
    outer_slug_od = property(get_outer_slug_od, None, None, None)

    def get_outer_slug_id(self):
        return self.__get_outer_slug_id()
    outer_slug_id = property(get_outer_slug_id, None, None, None)

    def get_inner_slug_od(self):
        return self.__get_inner_slug_od()
    inner_slug_od = property(get_inner_slug_od, None, None, None)

    def get_inner_slug_id(self):
        return self.__get_inner_slug_id()
    inner_slug_id = property(get_inner_slug_id, None, None, None)

    def get_cladding_wall_thickness(self):
        return self.__get_cladding_wall_thickness()
    cladding_wall_thickness = property(
        get_cladding_wall_thickness, None, None, None)

    def get_cladding_end_thickness(self):
        return self.__get_cladding_end_thickness()
    cladding_end_thickness = property(get_cladding_end_thickness, None, None, None)

    def get_slug_fuel_volume(self): return self.__get_slug_fuel_volume()
    slug_fuel_volume = property(get_slug_fuel_volume, None, None, None)

    def get_fuel_volume(self):
        return self.__get_fuel_volume()
    fuel_volume = property(get_fuel_volume, None, None, None)

    def get_slug_cladding_volume(self):
        return self.__get_slug_cladding_volume()
    slug_cladding_volume = property(get_slug_cladding_volume, None, None, None)

    def get_cladding_volume(self):
        return self.__get_cladding_volume()
    cladding_volume = property(get_cladding_volume, None, None, None)

    def get_fuel_mass(self):
        return self.__fuel_phase.GetValue('mass') # mass of the solid phase
    fuel_mass = property(get_fuel_mass, None, None, None)

    def get_fuel_mass_unit(self):
        return self.__fuel_phase.GetQuantity('mass').unit # mass of the solid phase
    fuel_mass_unit = property(get_fuel_mass_unit, None, None, None)

    def get_radioactivity(self):   # radioactivity of the fuel bucket (fix me)
        return self.__fuel_phase.GetValue('radioactivity')
    radioactivity = property(get_radioactivity, None, None, None)

    def get_gamma_pwr(self):
        return self.__fuel_phase.GetValue('gamma')   # gamma pwr of the fuel bucket
    gamma_pwr = property(get_gamma_pwr, None, None, None)

    def get_heat_pwr(self):
        return self.__fuel_phase.GetValue('heat')   # heat pwr of the fuel bucket
    heat_pwr = property(get_heat_pwr, None, None, None)

    def get_fuel_radioactivity(self):         # radioactivity of the solid phase
        return self.__fuel_phase.GetValue('radioactivity')
    fuel_radioactivity = property(get_fuel_radioactivity, None, None, None)

    def get_fuel_phase(self):
        return self.__fuel_phase

    def set_fuel_phase(self, phase):
        self.__fuel_phase = deepcopy(phase)
    fuel_phase = property(get_fuel_phase, set_fuel_phase, None, None)

    def get_cladding_phase(self):
        return self.__cladding_phase

    def set_cladding_phase(self, phase):
        self.__cladding_phase = deepcopy(phase)
    cladding_phase = property(get_cladding_phase, set_cladding_phase, None, None)

# *********************************************************************************
# Private helper functions (internal use: __)

    def __get_n_slugs(self):
        return int(self.__specs.loc['Number of slugs', 1])

    def __get_fuel_enrichment(self):
        return float(self.__specs.loc['Enrichment [U-235 wt%]', 1])

    def __get_outer_slug_fresh_u_mass(self):
        return float(
            self.__specs.loc['U mass outer slug [kg]', 1]) * 1000.0  # [g]

    def __get_inner_slug_fresh_u_mass(self):
        return float(
            self.__specs.loc['U mass inner slug [kg]', 1]) * 1000.0  # [g]

    def __get_outer_slug_cladding_mass(self):
        return float(
            self.__specs.loc['Cladding mass outer slug [kg]', 1]) * 1000.0  # [g]

    def __get_inner_slug_cladding_mass(self):
        return float(
            self.__specs.loc['Cladding mass inner slug [kg]', 1]) * 1000.0  # [g]

    def __get_fresh_u_mass(self):
        n_slugs = self.__get_n_slugs()
        uMassOuterSlug = self.__get_outer_slug_fresh_u_mass()
        uMassInnerSlug = self.__get_inner_slug_fresh_u_mass()
        return n_slugs * (uMassOuterSlug + uMassInnerSlug)

    def __get_fresh_u238_mass(self):
        totalUMass = self.__get_fresh_u_mass()
        fuelEnrichment = self.__get_fuel_enrichment()
        return totalUMass * (1.0 - fuelEnrichment / 100.0)

    def __get_fresh_u235_mass(self):
        totalUMass = self.__get_fresh_u_mass()
        fuelEnrichment = self.__get_fuel_enrichment()
        return totalUMass * fuelEnrichment / 100.0

    def __get_cladding_mass(self):
        n_slugs = self.__get_n_slugs()
        cladMassOuterSlug = self.__get_outer_slug_cladding_mass()
        cladMassInnerSlug = self.__get_inner_slug_cladding_mass()
        return n_slugs * (cladMassOuterSlug + cladMassInnerSlug)

    def __get_slug_length(self):
        return float(self.__specs.loc['Slug length [in]', 1]) * 2.54  # cm

    def __set_slug_length(self, x):
        self.__specs.loc['Slug length [in]', 1] = x / 2.54  # in
        return

    def __get_outer_slug_od(self):
        return float(self.__specs.loc['Outer slug O.D. [in]', 1]) * 2.54  # cm

    def __get_outer_slug_id(self):
        return float(self.__specs.loc['Outer slug I.D. [in]', 1]) * 2.54  # cm

    def __get_inner_slug_od(self):
        return float(self.__specs.loc['Inner slug O.D. [in]', 1]) * 2.54  # cm

    def __get_inner_slug_id(self):
        return float(self.__specs.loc['Inner slug I.D. [in]', 1]) * 2.54  # cm

    def __get_cladding_wall_thickness(self):
        return float(
            self.__specs.loc['Cladding wall thickness [mm]', 1]) / 10.0  # cm

    def __get_cladding_end_thickness(self):
        return float(
            self.__specs.loc['Cladding end cap thickness [mm]', 1]) / 10.0  # cm

    def __get_slug_fuel_volume(self):
        slugLength = self.__get_slug_length()
        cladWallThickness = self.__get_cladding_wall_thickness()
        cladEndThickness = self.__get_cladding_end_thickness()
        fuelLength = slugLength - 2.0 * cladEndThickness
        fuelOuterSlugOuterRadius = self.__get_outer_slug_od() / 2.0 - cladWallThickness
        fuelOuterSlugInnerRadius = self.__get_outer_slug_id() / 2.0 + cladWallThickness
        outerVolume = fuelLength * math.pi * \
            (fuelOuterSlugOuterRadius**2 - fuelOuterSlugInnerRadius**2)
        fuelInnerSlugOuterRadius = self.__get_inner_slug_od() / 2.0 - cladWallThickness
        fuelInnerSlugInnerRadius = self.__get_inner_slug_id() / 2.0 + cladWallThickness
        innerVolume = fuelLength * math.pi * \
            (fuelInnerSlugOuterRadius**2 - fuelInnerSlugInnerRadius**2)
        return outerVolume + innerVolume

    def __get_slug_volume(self):
        slugLength = self.__get_slug_length()
        outerSlugOuterRadius = self.__get_outer_slug_od() / 2.0
        outerSlugInnerRadius = self.__get_outer_slug_id() / 2.0
        outerVolume = slugLength * math.pi * \
            (outerSlugOuterRadius**2 - outerSlugInnerRadius**2)
        innerSlugOuterRadius = self.__get_inner_slug_od() / 2.0
        innerSlugInnerRadius = self.__get_inner_slug_id() / 2.0
        innerVolume = slugLength * math.pi * \
            (innerSlugOuterRadius**2 - innerSlugInnerRadius**2)
        return outerVolume + innerVolume

    def __get_slug_cladding_volume(self):
        return self.__get_slug_volume() - self.__get_slug_fuel_volume()

    def __get_fuel_volume(self):
        slugFuelVolume = self.__get_slug_fuel_volume()
        nFuelSlugs = self.__get_n_slugs()
        return slugFuelVolume * nFuelSlugs

    def __get_cladding_volume(self):
        slugCladdingVolume = self.__get_slug_cladding_volume()
        nFuelSlugs = self.__get_n_slugs()
        return slugCladdingVolume * nFuelSlugs

#*************************************************************************
# Printing of data members
    def __str__(self):
        s = '\nFuelBucket():\n\t*******\n\t specs:\n\t*******\n\t %s\n'
        t = '\t************\n\t fuel_phase:\n\t**********\n\t %s\n'
        u = '\t***************\n\t cladding_phase:\n\t*****************\n\t %s\n'
        stu = s + t + u
        return stu % (self.__specs, self.__fuel_phase, self.__cladding_phase)

    def __repr__(self):
        s = '\nFuelBucket():\n\t*******\n\t specs:\n\t*******\n\t %s\n'
        t = '\t************\n\t fuel_phase:\n\t**********\n\t %s\n'
        u = '\t****************\n\t cladding_phase:\n\t****************\n\t %s\n'
        stu = s + t + u
        return stu % (self.__specs, self.__fuel_phase, self.__cladding_phase)

# ====================== end class FuelBucket: ===================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Cortix: a program for system-level modules coupling, execution, and analysis.

Cortix is a library and it is used by means of a driver. This file is a simple example
of a driver. Many Cortix objects can be ran simultaneously; a single object
may be sufficient since many simulation/tasks can be ran via one object.

As Cortix evolves additional complexity may be added to this driver and/or other
driver examples can be created.
"""
# *********************************************************************************
import os
from mpi4py import MPI
from cortix import Cortix
# *********************************************************************************


def main():

    pwd = os.path.dirname(__file__)
    full_path_config_file = os.path.join(pwd, '../input/cortix-config.xml')

    # NB: if another instantiation of Cortix occurs, the cortix wrk directory specified
    #     in the cortix configuration file must be different, else the logging facility
    #     will have log file collision.

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()

    if (rank == 0):
        cortix1 = Cortix('cortix1', full_path_config_file)
        cortix1.run_simulations( task_name='solo-pyplot')  # see cortix-config.xml

#  cortix1 = Cortix( 'cortix1', full_path_config_file )
#  cortix1.run_simulations( task_name='solo-fueldepot' ) # see cortix-config.xml
# ---------------------- end def main():----------------------------------


# *********************************************************************************
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Application class for Cortix.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import sys
import logging
from cortix.src.utils.xmltree import XMLTree
from cortix.src.utils.set_logger_level import set_logger_level

from cortix.src.module  import Module
from cortix.src.network import Network
#*********************************************************************************

class Application:
    r'''
    An Application is a singleton class composed of Module objects, and Network
    objects; the latter involve Module objects in various combinations. Each
    combination is assigned to a Network object.
    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__(self, app_work_dir, config_xml_tree):

        assert isinstance(app_work_dir, str), '-> app_work_dir is invalid'
        assert os.path.isdir(app_work_dir), 'Work directory not available.'

        assert isinstance(config_xml_tree, XMLTree), '-> config_xml_tree invalid'
        assert config_xml_tree.get_node_tag() == 'application'

        # Read the application name
        self.__name = config_xml_tree.get_node_attribute('name')

        # Set the work directory (previously created)
        self.__work_dir = app_work_dir

        # Set the module library for the whole application
        node = config_xml_tree.get_sub_node('module_library')
        assert node.get_node_tag() == 'module_library', 'FATAL.'

        self.__importlib_name = node.get_node_attribute('importlib_name')

        for child in node.get_node_children():
            (elem, tag, attributes, text) = child
            if tag == 'home_dir':
                self.__module_lib_full_home_dir = text.strip()

        if self.__module_lib_full_home_dir[-1] == '/':
            self.__module_lib_full_home_dir.strip('/')

        # Add library full path to python module search
        sys.path.insert(1, self.__module_lib_full_home_dir)

        # Create logging facility
        self.__create_logging_facility( config_xml_tree )

        #==============
        # Setup modules
        #==============
        self.__modules = list()

        self.__setup_modules(config_xml_tree)

        #===============
        # Setup networks 
        #===============
        self.__networks = list()

        self.__setup_networks(config_xml_tree)

        self.__log.info('Created application: %s', self.__name)

        return

    def __del__(self):

        self.__log.info('destroyed application: %s', self.__name)

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def __get_networks(self):
        '''
        `list(str)`:List of names of network objects
        '''

        return self.__networks

    networks = property(__get_networks, None, None, None)

    def get_network(self, name):
        '''
        Returns a network with a given name. None if the name doesn't exist.

        Parameters
        ----------
        name: str

        Returns
        -------
        net: cortix.network or None
           Default: None
        '''

        for net in self.__networks:
            if net.name == name:
                return net

        return None

    def __get_modules(self):
        '''
        `list(str)`:List of names of Cortix module objects
        '''

        return self.__modules

    modules = property(__get_modules, None, None, None)

    def get_module(self, name):
        """
        Returns a module with a given name.  None if the name doesn't exist.
        """
        for mod in self.__modules:
            if mod.name == name:
                return mod
        return None

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __create_logging_facility(self, config_xml_tree):
        '''
        A helper function to setup the logging facility used in self.__init__().
        '''

        assert config_xml_tree.get_node_tag() == 'application'

        # Create the logging facility for the singleton object
        node = config_xml_tree.get_sub_node('logger')

        logger_name = 'app:'+self.__name # prefix to avoid clash of loggers
        self.__log = logging.getLogger(logger_name)
        self.__log.setLevel(logging.NOTSET)

        logger_level = node.get_node_attribute('level')
        self.__log = set_logger_level(self.__log, logger_name, logger_level)

        file_handler = logging.FileHandler(self.__work_dir + 'app.log')
        file_handler.setLevel(logging.NOTSET)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.NOTSET)

        for child in node.get_node_children():
            (elem,tag,attributes,text) = child
            elem = XMLTree( elem )
            if tag == 'file_handler':
                file_handler_level = elem.get_node_attribute('level')
                file_handler = set_logger_level(file_handler, logger_name,
                                                file_handler_level)
            if tag == 'console_handler':
                console_handler_level = elem.get_node_attribute('level')
                console_handler = set_logger_level(console_handler, logger_name,
                                                   console_handler_level)

        # formatter added to handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        # add handlers to logger
        self.__log.addHandler(file_handler)
        self.__log.addHandler(console_handler)
        self.__log.info('created Application logger: %s', self.__name)
        self.__log.debug('logger level: %s', logger_level)
        self.__log.debug('logger file handler level: %s', file_handler_level)
        self.__log.debug('logger console handler level: %s', console_handler_level)

        return

    def __setup_modules( self, config_xml_tree ):
        '''
        A helper function used by the Application constructor to setup the modules
        portion of the Application. What this does is to go into the Cortix config file
        and read all modules that part of the module tag whether they are used in a
        given task or not.
        '''

        self.__log.debug('Start __setup_modules()')

        for mod_config_xml_node in config_xml_tree.get_all_sub_nodes('module'):

            assert mod_config_xml_node.get_node_tag() == 'module'

            # Modules log into the Application logger because they may or may not be 
            # used in a task. Running tasks will effectively tell what modules are used.
            new_module = Module( self.__log, self.__importlib_name,
                                 self.__module_lib_full_home_dir,
                                 mod_config_xml_node )

            # check for a duplicate module before appending a new one
            for module in self.__modules:
                mod_lib_dir_name   = module.library_home_dir
                mod_importlib_name = module.importlib_name

                if new_module.name == module.name:
                    assert new_module.library_home_dir != mod_lib_dir_name,\
                            ' duplicate module; ABORT.'
                    assert new_module.importlib_name != mod_importlib_name,\
                            ' duplicate module; ABORT.'

            # add module to list
            self.__modules.append( new_module )

            self.__log.debug('appended module %s',
                    mod_config_xml_node.get_node_attribute('name'))

        self.__log.debug('end __setup_modules()')

        return

    def __setup_networks( self, config_xml_tree ):
        '''
        A helper function used by the Application constructor to setup the networks
        portion of the Application.
        '''

        self.__log.debug('start __setup_networks()')

        for net_config_xml_node in config_xml_tree.get_all_sub_nodes('network'):

            assert net_config_xml_node.get_node_tag() == 'network'

            network = Network( net_config_xml_node )

            self.__networks.append(network)

            self.__log.debug('appended network %s',
                    net_config_xml_node.get_node_attribute('name'))

        self.__log.debug('end __setup_networks()')

        return

#======================= end class Application: ==================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Launcher functionality of the Cortix Class.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import logging
import time
import datetime
from threading import Thread
import importlib
import xml.etree.ElementTree as ElementTree
import concurrent.futures
#*********************************************************************************

class Launcher(Thread):
    '''
    The Launcher class handles the main funcitonality of stepping through the
    simulation time, and monitoring its progress.
    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__(self, 
            importlib_name, mod_lib_home_dir, module_name, 
            slot_id,
            input_full_path_file_name,
            manifest_full_path_file_name,
            work_dir,
            cortix_param_full_path_file_name,
            cortix_comm_full_path_file_name,
            runtime_status_full_path):

        assert cortix_param_full_path_file_name[-1] is not '/', \
                '%r'%cortix_param_full_path_file_name
        assert cortix_comm_full_path_file_name[-1] is not '/' \
                '%r'%cortix_comm_full_path_file_name

        cortix_path = os.path.abspath(os.path.join(__file__, '../../..'))

        self.__input_full_path_file_name = input_full_path_file_name

        self.__module_name = module_name
        self.__slot_id = slot_id
        self.__cortix_param_full_path_file_name = cortix_param_full_path_file_name
        self.__cortix_comm_full_path_file_name = cortix_comm_full_path_file_name
        self.__runtime_status_full_path = runtime_status_full_path
        self.__manifest_full_path_file_name = manifest_full_path_file_name
        self.__work_dir = work_dir

        # Create logging facility
        self.__create_logging_facility()

        lib_module_driver = importlib_name + '.' + module_name + '.cortix_driver'
        self.__log.info('try importing module driver: %s', lib_module_driver)

        # import a guest Cortix module through its driver
        try:
            self.__py_module = importlib.import_module( lib_module_driver )
        except Exception as error:
            log.error('importlib error: ', error)

        self.__log.info('imported pyModule: %s', str(self.__py_module))

        # Launcher thread initialiation
        super(Launcher, self).__init__()

        return

    def __del__(self):

        self.__log.info('destroyed launcher-%s',
                      self.__module_name + '_' + str(self.__slot_id))

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def run(self):
        '''
        Function used to timestep through the modules.
        Runs the simulation from start to end, and monitors
        its progress at each time step.
        '''

        # log info           
        self.__log.info('entered run() %s', self.__module_name +
                      '_' + str(self.__slot_id) )

        # Verify the module input file name with full path.
        # This input file may be empty or used by this driver and/or the
        # native/wrapped module.
        assert os.path.isfile(self.__input_full_path_file_name), \
            'file %r not available;stop.' % self.__input_full_path_file_name

        # Read the Cortix parameter file: cortix-param.xml
        # cortix_param_full_path_file_name
        assert os.path.isfile(self.__cortix_param_full_path_file_name), \
            'file %r not available;stop.' % self.__cortix_param_full_path_file_name

        # For now Cortix advances in unit of minutes; change this in the future
        tree = ElementTree.parse(self.__cortix_param_full_path_file_name)
        cortix_param_xml_root_node = tree.getroot()
        node = cortix_param_xml_root_node.find('start_time')
        cortix_start_time_unit = node.get('unit')
        cortix_start_time = float(node.text.strip())

        if cortix_start_time_unit == 'minute':
            cortix_start_time *= 1.0
        elif cortix_start_time_unit == 'hour':
            cortix_start_time *= 60.0
        elif cortix_start_time_unit == 'day':
            cortix_start_time *= 24.0 * 60.0
        else:
            assert False, 'time unit invalid: %r' % (cortix_start_time_unit)

        node = cortix_param_xml_root_node.find('evolve_time')
        evolve_time_unit = node.get('unit')
        evolve_time = float(node.text.strip())

        if evolve_time_unit == 'minute':
            evolve_time *= 1.0
        elif evolve_time_unit == 'hour':
            evolve_time *= 60.0
        elif evolve_time_unit == 'day':
            evolve_time *= 24.0 * 60.0
        else:
            assert False, 'time unit invalid: %r' % (evolve_time_unit)

        node = cortix_param_xml_root_node.find('time_step')
        cortix_time_step_unit = node.get('unit')
        cortix_time_step = float(node.text.strip())

        if cortix_time_step_unit == 'minute':
            cortix_time_step *= 1.0
        elif cortix_time_step_unit == 'hour':
            cortix_time_step *= 60.0
        elif cortix_time_step_unit == 'day':
            cortix_time_step *= 24.0 * 60.0
        elif cortix_time_step_unit == 'second':
            cortix_time_step /= 60.0
        else:
            assert False, 'time unit invalid: %r' % (cortix_time_step_unit)

        cortix_time_unit = 'minute'

        # collect information from the Cortix communication file for this guest module
        assert os.path.isfile(self.__cortix_comm_full_path_file_name),\
            'file %r not available;stop.' % self.__cortix_comm_full_path_file_name

        tree = ElementTree.parse(self.__cortix_comm_full_path_file_name)
        cortix_comm_xml_root_node = tree.getroot()

        # Setup ports
        nodes = cortix_comm_xml_root_node.findall('port')
        ports = list()
        if nodes is not None:
            for node in nodes:
                port_name = node.get('name')
                port_type = node.get('type')
                port_file = node.get('file')
                port_directory = node.get('directory')

                if port_file is not None:
                    ports.append((port_name, port_type, port_file))
                elif port_directory is not None:
                    ports.append((port_name, port_type, port_directory))
                else:
                    assert False, 'port mode incorrect. fatal.'

        tree = None
        self.__log.debug('ports: %s', str(ports))

        # Add evolve time to start time  
        cortix_final_time = cortix_start_time + evolve_time

        # Create the guest code driver
        guest_driver = self.__py_module.CortixDriver( self.__slot_id,
                                                      self.__input_full_path_file_name,
                                                      self.__manifest_full_path_file_name,
                                                      self.__work_dir,
                                                      ports,
                                                      cortix_start_time,
                                                      cortix_final_time,
                                                      cortix_time_step,
                                                      cortix_time_unit
                                                    )

        s = 'guest_driver = CortixDriver( slot_id=' + str(self.__slot_id) + \
            ', input file=' + self.__input_full_path_file_name + \
            ', exec file=' + self.__manifest_full_path_file_name + \
            ', work dir=' + self.__work_dir + \
            ', ports=' + str(ports) + \
            ', cortix_start_time=' + str(cortix_start_time) + \
            ', cortix_final_time=' + str(cortix_final_time) + \
            ', cortix_time_unit=' + cortix_time_unit + \
            ', cortix_time_step=' + str(cortix_time_step) + \
            ', cortix_time_step_unit=', cortix_time_step_unit
        self.__log.info(s)

        # Evolve the module
        self.__set_runtime_status('running')
        self.__log.info("__set_runtime_status(self, 'running')")

        cortix_time = cortix_start_time

        before_final_time = True

        while cortix_time <= cortix_final_time or before_final_time:

            if cortix_time >= cortix_final_time:  # make sure the final time is reached
               before_final_time = False          # or exceeded by a epsilon amount

            s = ''
            self.__log.debug(s)
            s = '**************************************************************' + \
                '**************'
            self.__log.debug(s)
            s = 'CORTIX::LAUNCHER->***->LAUNCHER->***->LAUNCHER->***->' + \
                'LAUNCHER->***->LAUNCHER'
            self.__log.debug(s)
            s = '**************************************************************' + \
                '**************'
            self.__log.debug(s)

            s = 'run(' + str(round(cortix_time, 3)) + '[min]): ' # todo: change time unit
            self.__log.debug(s)
            s = 'run(' + str(round(cortix_time, 3)) + '[min]) ' # todo: change time unit
            self.__log.info(s)

            start_time = time.time()

            # Data exchange at cortix_time (call ports first)
            guest_driver.call_ports( cortix_time )

            # Advance to cortix_time + cortix_time_step (call execute second)
            guest_driver.execute( cortix_time, cortix_time_step )

            end_time = time.time()
            s = 'CPU elapsed time (s): ' + str(round(end_time - start_time, 2))
            self.__log.debug(s)

            cortix_time += cortix_time_step

        self.__set_runtime_status('finished')
        self.__log.info("__set_runtime_status(self, 'finished'")

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __set_runtime_status(self, status):
        """
        Helper function used by the launcher constructor to output status of the
        current run.
        """

        status = status.strip()
        assert status == 'running' or status == 'finished', 'status invalid.'

        fout = open(self.__runtime_status_full_path, 'w')
        fout.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        fout.write('<!-- Written by Launcher::__set_runtime_status.py -->\n')

        today = datetime.datetime.today()
        fout.write('<!-- ' + str(today) + ' -->\n')
        fout.write('<runtime>\n')
        fout.write('<status>' + status + '</status>\n')
        fout.write('</runtime>\n')

        fout.close()

        return

    def __create_logging_facility(self):
        '''
        A helper function to setup the logging facility used in self.__init__()
        '''

        # Create logger for this launcher and its imported pymodule; Cortix modules
        # should use this for logging
        log = logging.getLogger('launcher-' + self.__module_name + '_' +
                                str(self.__slot_id))
        log.setLevel(logging.DEBUG)

        # Create file handler for logs; same place as the Cortix comm file for module
        i = self.__cortix_comm_full_path_file_name.rfind('/')
        directory = self.__cortix_comm_full_path_file_name[:i]
        full_path_launcher_dir = directory + '/'
        file_handle = logging.FileHandler(full_path_launcher_dir + 'launcher.log')
        file_handle.setLevel(logging.DEBUG)

        # Create console handler with a higher log level
        console_handle = logging.StreamHandler()
        console_handle.setLevel(logging.INFO)

        # Create formatter and add it to the handlers
        formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handle.setFormatter(formatter)
        console_handle.setFormatter(formatter)

        # add handles to the logger
        log.addHandler(file_handle)
        log.addHandler(console_handle)

        log.info('created logger')
        log.debug('input file: %s', self.__input_full_path_file_name)
        log.debug('param file: %s', self.__cortix_param_full_path_file_name)
        log.debug('comm file: %s', self.__cortix_comm_full_path_file_name)

        self.__log = log

        return

#======================= end class Launcher: =====================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
This file contains a helper function used by functions across the Cortix project
to set the level of the logger.
'''
# *********************************************************************************
import logging
# *********************************************************************************


def set_logger_level(handler, handler_name, handler_level):
    '''
    This is a helper function that takes in a file/console handler and sets its
    logger level accordingly.
    '''

    if handler_level == 'DEBUG':
        handler.setLevel(logging.DEBUG)
    elif handler_level == 'INFO':
        handler.setLevel(logging.INFO)
    elif handler_level == 'WARN':
        handler.setLevel(logging.WARN)
    elif handler_level == 'ERROR':
        handler.setLevel(logging.ERROR)
    elif handler_level == 'CRITICAL':
        handler.setLevel(logging.CRITICAL)
    elif handler_level == 'FATAL':
        handler.setLevel(logging.FATAL)
    else:
        assert False, 'File handler log level for %r: %r invalid'%\
                (handler_name, handler_level)

    return handler
# ---------------------- end def set_logger_level():----------------------



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Droplet module example in Cortix.
'''
#*********************************************************************************
import os, sys, io, time
import logging
from collections import namedtuple
import math
import numpy as npy

from cortix.src.utils.xmltree import XMLTree
from cortix.support.quantity import Quantity
from cortix.support.specie   import Specie
from cortix.support.phase    import Phase
#*********************************************************************************

class Droplet():
    r'''
    Droplet module used example in Cortix.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__( self,
            slot_id,
            input_full_path_file_name,
            manifest_full_path_file_name,
            work_dir,
            ports = list(),
            cortix_start_time = 0.0,
            cortix_final_time = 0.0,
            cortix_time_step = 0.0,
            cortix_time_unit=None
                ):

        #.............................................................................
        # Sanity test

        assert isinstance(slot_id, int), '-> slot_id type %r is invalid.' % type(slot_id)
        assert isinstance(ports, list), '-> ports type %r is invalid.'  % type(ports)
        assert len(ports) > 0
        assert isinstance(cortix_start_time,float), '-> time type %r is invalid.' % \
                type(cortix_start_time)
        assert isinstance(cortix_final_time, float), '-> time type %r is invalid.' % \
                type(cortix_final_time)
        assert isinstance(cortix_time_step, float), '-> time step type %r is invalid.' % \
                type(cortix_time_step)
        assert isinstance(cortix_time_unit, str), '-> time unit type %r is invalid.' % \
                type(cortix_time_unit)

        # Logging: access Cortix Launcher logging facility
        self.__log = logging.getLogger('launcher-droplet_'+str(slot_id)+'.cortix_driver.droplet')
        self.__log.info('initializing an object of Droplet()')

        # Read the manisfest 
        self.__read_manifest( manifest_full_path_file_name )
        self.__log.info(self.__port_diagram)

        #.............................................................................
        # Member data 

        self.__slot_id = slot_id
        self.__ports   = ports

        # Convert Cortix's time unit to Droplet's internal time unit
        if cortix_time_unit == 'minute':
            self.__time_unit_scale = 60.0
        elif cortix_time_unit == 'second':
            self.__time_unit_scale = 1.0
        elif cortix_time_unit == 'hour':
            self.__time_unit_scale = 60.0*60.0
        else:
            assert False, 'Cortix time_unit: %r not acceptable.' % time_unit

        self.__cortix_time_unit = cortix_time_unit

        self.__start_time = cortix_start_time * self.__time_unit_scale # Droplet time unit
        self.__final_time = cortix_final_time * self.__time_unit_scale # Droplet time unit

        if work_dir[-1] != '/': work_dir = work_dir + '/'
        self.__wrkDir = work_dir

        # Signal to start operation
        self.__goSignal = True     # start operation immediately
        for port in self.__ports:  # if there is a signal port, start operation accordingly
            (port_name, port_type, this_port_file) = port
            if port_name == 'go-signal' and port_type == 'use': self.__go_signal = False

        self.__setup_time = 60.0  # time unit; a delay time before starting to run

        #.............................................................................
        # Input ports
        # Read input information if any

        #fin = open(input_full_path_file_name,'r')

        # Configuration member data   

        self.__pyplot_scale = 'linear'
        #  self.__ode_integrator = 'scikits.odes' # or 'scipy.integrate' 
        self.__ode_integrator = 'scipy.integrate'

        # Domain specific member data 

        gravity = 9.81 # [m/s^2] acceleration of gravity

        Params = namedtuple('Params',['gravity'])
        self.__params = Params( gravity )

        # Setup the material phase as a liquid 
        species = list()

        water = Specie( name='water', formulaName='H2O(l)', phase='liquid', atoms=['2*H','O'] )
        water.massCCUnit = 'g/cc'
        water.molarCCUnit = 'mole/cc'

        species.append(water)

        quantities = list()

        # spatial position
        spatial_position = Quantity( name='spatial-position', formalName='Spatial Pos.', unit='m' )
        quantities.append( spatial_position )
        # velocity
        velocity = Quantity( name='velocity', formalName='Veloc.', unit='m/s' )
        quantities.append( velocity )

        self.__liquid_phase = Phase( self.__start_time, species=species, quantities=quantities)

        # Initialize phase
        water_mass_cc = 0.99965 # [g/cc]
        self.__liquid_phase.SetValue( 'water', water_mass_cc, self.__start_time )

        x_0 = (0.0,0.0,1000.0)  # initial height [m] above ground at 0
        self.__liquid_phase.SetValue( 'spatial-position', x_0, self.__start_time )

        v_0 = (0.0,0.0,0.0)     # initial velocity [m/s]
        self.__liquid_phase.SetValue( 'velocity', v_0, self.__start_time )

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def call_ports( self, cortix_time=0.0 ):
        '''
        Transfer data at cortix_time
        '''

        cortix_time *= self.__time_unit_scale  # convert to Droplet time unit

        # provide data to all provide ports 
        self.__provide_data( provide_port_name='output', at_time=cortix_time )
        self.__provide_data( provide_port_name='state',  at_time=cortix_time )

        # use data using the 'use-port-name' of the module
        #self.__use_data( use_port_name='use-port-name', at_time=cortix_time )

        return

    def execute( self, cortix_time=0.0, cortix_time_step=0.0 ):
        '''
        Evolve system from cortix_time to cortix_time + cortix_time_step
        '''

        cortix_time      *= self.__time_unit_scale  # convert to Droplet time unit
        cortix_time_step *= self.__time_unit_scale  # convert to Droplet time unit

        s = 'execute('+str(round(cortix_time,2))+'[min]): '
        self.__log.debug(s)

        self.__evolve( cortix_time, cortix_time_step )

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __provide_data( self, provide_port_name=None, at_time=0.0 ):

        # Access the port file
        port_file = self.__get_port_file( provide_port_name = provide_port_name )

        # Provide data to port files
        if provide_port_name == 'output' and port_file is not None:
            self.__provide_output( port_file, at_time )

        if provide_port_name == 'state' and port_file is not None:
             self.__provide_state( port_file, at_time )

        return

    def __use_data( self, use_port_name=None, at_time=0.0 ):

        # Access the port file
        port_file = self.__get_port_file( use_port_name = use_port_name )

        # Use data from port file
        #if use_port_name == 'use-port-name' and port_file is not None:  
        #   self.__use_mymodule_method( port_file, at_time )

        return

    def __get_port_file( self, use_port_name=None, provide_port_name=None ):
        '''
        This may return a None port_file
        '''

        port_file = None

        #..........
        # Use ports
        #..........
        if use_port_name is not None:

            assert provide_port_name is None

            for port in self.__ports:

                (port_name,port_type,this_port_file) = port

                if port_name == use_port_name and port_type == 'use':
                    port_file = this_port_file

            if port_file is None:
                return None

            max_n_trials = 50
            n_trials     = 0
            while os.path.isfile(port_file) is False and n_trials <= max_n_trials:
                n_trials += 1
                time.sleep(0.1)

            if n_trials > max_n_trials:
                s = '__get_port_file(): waited ' + str(n_trials) + ' trials for port: ' +\
                        port_file
                self.__log.warn(s)

            assert os.path.isfile(port_file) is True,'port_file %r not available; stop.'\
                    % port_file

        #..............
        # Provide ports
        #..............
        if provide_port_name is not None:

            assert use_port_name is None

            for port in self.__ports:
                (port_name,port_type,this_port_file) = port
                if port_name == provide_port_name and port_type == 'provide': \
                        port_file = this_port_file

        return port_file

    def __provide_output( self, port_file, at_time ):
        '''
        Provide data that will remain in disk after the simulation ends.
        '''
        import datetime

        # if the first time step, write the header of a table data file
        if at_time == self.__start_time:

            assert os.path.isfile(port_file) is False, 'port_file %r exists; stop.' % port_file
            fout = open(port_file,'w')

            fout.write('# name:   '+'droplet_'+str(self.__slot_id)); fout.write('\n')
            fout.write('# author: '+'cortix.examples.modulib.droplet'); fout.write('\n')
            fout.write('# version:'+'0.1'); fout.write('\n')
            today = datetime.datetime.today()
            fout.write('# today:  '+str(today)); fout.write('\n')
            fout.write('#')
            # write file header
            fout.write('%17s'%('Time[sec]'))
            # mass density   
            for specie in self.__liquid_phase.GetSpecies():
              fout.write('%18s'%(specie.formulaName+'['+specie.massCCUnit+']'))
            # quantities     
            for quant in self.__liquid_phase.GetQuantities():
                if quant.name == 'spatial-position' or quant.name == 'velocity':
                   for i in range(3):
                       fout.write('%18s'%(quant.formalName+str(i+1)+'['+quant.unit+']'))
                else:
                   fout.write('%18s'%(quant.formalName+'['+quant.unit+']'))

            fout.write('\n')
            fout.close()

        # write history data

        fout = open(port_file,'a')

        # Droplet time 
        fout.write('%18.6e' % (at_time))

        # mass density   
        for specie in self.__liquid_phase.GetSpecies():
            rho = self.__liquid_phase.GetValue(specie.name, at_time)
            fout.write('%18.6e'%(rho))
        # quantities     
        for quant in self.__liquid_phase.GetQuantities():
            val = self.__liquid_phase.GetValue(quant.name, at_time)
            if quant.name == 'spatial-position' or quant.name == 'velocity':
                for v in val:
                    fout.write('%18.6e'%(v))
            else:
                fout.write('%18.6e'%(val))

        fout.write('\n')
        fout.close()

        return

    def __provide_state( self, port_file, at_time ):
        '''
        Provide the internal state of the module. This is typically used by another
        module such as '.cortix.modulib.pyplot'; this XML file type is now deprecated.
        However to have a dynamic plotting option, create this history file in the
        time unit of Cortix; that is, undo the time scaling.
        '''
        import datetime
        import xml.etree.ElementTree as ElementTree
        from threading import Lock

        n_digits_precision = 8

        # Write header
        if at_time == self.__start_time:

         assert os.path.isfile(port_file) is False, 'port_file %r exists; stop.'\
                 % port_file

         a = ElementTree.Element('time-sequence')
         a.set('name','droplet_'+str(self.__slot_id)+'-state')

         b = ElementTree.SubElement(a,'comment')
         b.set('author','cortix.examples.modulib.droplet')
         b.set('version','0.1')

         b = ElementTree.SubElement(a,'comment')
         today = datetime.datetime.today()
         b.set('today',str(today))

         b = ElementTree.SubElement(a,'time')
         b.set('unit',self.__cortix_time_unit)

         # setup the headers
         for specie in self.__liquid_phase.species:
             b = ElementTree.SubElement(a,'var')
             formula_name = specie.formulaName
             b.set('name',formula_name)
             unit = specie.massCCUnit
             b.set('unit',unit)
             b.set('legend','Droplet_'+str(self.__slot_id)+'-state')
             b.set('scale',self.__pyplot_scale)

         for quant in self.__liquid_phase.quantities:
             formal_name = quant.formalName
             if quant.name == 'spatial-position' or quant.name == 'velocity':
                 for i in range(3):
                     b = ElementTree.SubElement(a,'var')
                     b.set('name',formal_name+' '+str(i+1))
                     unit = quant.unit
                     b.set('unit',unit)
                     b.set('legend','Droplet_'+str(self.__slot_id)+'-state')
                     b.set('scale',self.__pyplot_scale)
             else:
                 b = ElementTree.SubElement(a,'var')
                 b.set('name',formal_name)
                 unit = quant.unit
                 b.set('unit',unit)
                 b.set('legend','Droplet_'+str(self.__slot_id)+'-state')
                 b.set('scale',self.__pyplot_scale)

         # Write values for all variables
         b = ElementTree.SubElement(a,'timeStamp')
         b.set('value',str(round(at_time/self.__time_unit_scale,n_digits_precision)))

         values = list()

         for specie in self.__liquid_phase.species:
             val = self.__liquid_phase.GetValue( specie.name, at_time )
             values.append( val )

         for quant in self.__liquid_phase.quantities:
             val = self.__liquid_phase.GetValue( quant.name, at_time )
             if quant.name == 'spatial-position' or quant.name == 'velocity':
                for v in val:
                    values.append( math.fabs(v) ) # pyplot can't take negative values
             else:
                values.append( val )

         # Flush out data
         text = str()
         for value in values:
             text += str(round(value,n_digits_precision)) + ','

         text = text[:-1]

         b.text = text

         tree = ElementTree.ElementTree(a)

         tree.write( port_file, xml_declaration=True, encoding="unicode", method="xml" )

        #-------------------------------------------------------------------------
        # if not the first time step then parse the existing history file and append 
        # to it
        else:

            mutex = Lock()
            mutex.acquire()

            tree = ElementTree.parse( port_file )
            root_node = tree.getroot()

            a = ElementTree.Element('timeStamp')
            a.set('value',str(round(at_time/self.__time_unit_scale,n_digits_precision)))

            # All variables values
            values = list()

            for specie in self.__liquid_phase.species:
                val = self.__liquid_phase.GetValue( specie.name, at_time )
                values.append( val )
            for quant in self.__liquid_phase.quantities:
                val = self.__liquid_phase.GetValue( quant.name, at_time )
                if quant.name == 'spatial-position' or quant.name == 'velocity':
                    for v in val:
                        values.append( math.fabs(v) ) # pyplot can't take negative values
                else:
                    values.append( val )

            # Flush out data
            text = str()
            for value in values:
                text += str(round(value,n_digits_precision)) + ','

            text = text[:-1]

            a.text = text

            root_node.append(a)

            tree.write( port_file, xml_declaration=True, encoding="unicode", method="xml" )

            mutex.release()

        return

    def __evolve( self, cortix_time=0.0, cortix_time_step=0.0 ):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`, :math:`u_1(0) = x_0`,
        :math:`u_2(0) = v_0 = \dot{u}_1(0)`,
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.
        When :math:`u_1(t)` is negative, bounce the droplet to a random height between
        0 and :math:`1.2\,x_0` with no velocity, and continue the time integration until
        :math:`t = t_f`.
        '''
        if self.__ode_integrator == 'scikits.odes':
            from scikits.odes import ode        # this requires the SUNDIALS ODE package 
        elif self.__ode_integrator == 'scipy.integrate':
            from scipy.integrate import odeint  # this ships with scipy 
        else:
            assert False, 'Fatal: invalid ode integrator config. %r'%self.__ode_integrator

        x_0 = self.__liquid_phase.GetValue( 'spatial-position', cortix_time )
        v_0 = self.__liquid_phase.GetValue( 'velocity', cortix_time )

        u_vec_0 = x_0 + v_0 # concatenate tuples

        # rhs function
        if self.__ode_integrator == 'scikits.odes':
            def rhs_fn(t, u_vec, dt_u_vec, params):
                dt_u_vec[0] = u_vec[3]              #  d_t u_1 = u_4
                dt_u_vec[3] = 0.0                   #  d_t u_4 = 0 

                dt_u_vec[1] = u_vec[4]              #  d_t u_2 = u_5
                dt_u_vec[4] = 0.0                   #  d_t u_5 = 0 

                dt_u_vec[2] = u_vec[5]              #  d_t u_3 = u_6
                dt_u_vec[5] = - params.gravity      #  d_t u_6 = -g
                return
        elif self.__ode_integrator == 'scipy.integrate':
            def rhs_fn(u_vec, t, gravity):
                dt_u_0 = u_vec[3]              #  d_t u_1 = u_4
                dt_u_3 = 0.0                   #  d_t u_4 = 0

                dt_u_1 = u_vec[4]              #  d_t u_2 = u_5
                dt_u_4 = 0.0                   #  d_t u_5 = 0 

                dt_u_2 = u_vec[5]              #  d_t u_3 = u_6
                dt_u_5 = - gravity             #  d_t u_6 = -g
                return [dt_u_0, dt_u_1, dt_u_2, dt_u_3, dt_u_4, dt_u_5]
        else:
            assert False, 'Fatal: invalid ode integrator config. %r'%self.__ode_integrator

        t_interval_sec = npy.linspace(0.0, cortix_time_step, num=2)

        if self.__ode_integrator == 'scikits.odes':
            cvode = ode('cvode', rhs_fn, user_data=self.__params, old_api=False)
            solution = cvode.solve( t_interval_sec, u_vec_0 )  # solve for time interval 

            results = solution.values
            message = solution.message

            assert solution.message == 'Successful function return.'
            assert solution.errors.t is None, 'errors.t = %r'%solution.errors.t
            assert solution.errors.y is None, 'errors.y = %r'%solution.errors.y

            u_vec = results.y[1,:] # solution vector at final time step

        elif self.__ode_integrator == 'scipy.integrate':
            ( u_vec_hist, info_dict ) = odeint( rhs_fn,
                                                u_vec_0, t_interval_sec,
                                                args=( self.__params.gravity, ),
                                                full_output=True
                                              )

            assert info_dict['message']=='Integration successful.',\
                   'Fatal: scipy.integrate.odeint failed %r'%info_dict['message']

            u_vec = u_vec_hist[1,:]  # solution vector at final time step
        else:
            assert False, 'Fatal: invalid ode integrator config. %r'%self.__ode_integrator


        values = self.__liquid_phase.GetRow( cortix_time ) # values at previous time

        at_time = cortix_time + cortix_time_step

        self.__liquid_phase.AddRow( at_time, values ) # repeat values for current time

        if u_vec[2] <= 0.0: # ground impact bounces the drop to a different height near original
            position = self.__liquid_phase.GetValue( 'spatial-position', self.__start_time )
            bounced_position = list(position)
            bounced_position[2] *= npy.random.random(1) * 1.2
            u_vec[0:3] = bounced_position[:]
            u_vec[3:]  = 0.0

        self.__liquid_phase.SetValue( 'spatial-position', tuple(u_vec[0:3]), at_time ) # update current values
        self.__liquid_phase.SetValue( 'velocity', tuple(u_vec[3:]), at_time ) # update current values

        return

    def __read_manifest( self, xml_tree_file ):
        '''
        Parse the manifest
        '''

        assert isinstance(xml_tree_file, str)

        # Read the manifest 
        xml_tree = XMLTree( xml_tree_file=xml_tree_file )

        assert xml_tree.get_node_tag() == 'module_manifest'

        assert xml_tree.get_node_attribute('name') == 'droplet'

        # List of (port_name, port_type, port_mode, port_multiplicity)
        __ports = list()

        self.__port_diagram = 'null-module-port-diagram'

        # Get manifest data  
        for child in xml_tree.get_node_children():
            (elem, tag, attributes, text) = child

            if tag == 'port':

                text = text.strip()

                assert len(attributes) == 3, "only <= 3 attributes allowed."

                tmp = dict()  # store port name and three attributes

                for attribute in attributes:
                    key = attribute[0].strip()
                    val = attribute[1].strip()

                    if key == 'type':
                        assert val == 'use' or val == 'provide' or val == 'input' or \
                            val == 'output', 'port attribute value invalid.'
                        tmp['port_name'] = text  # port_name
                        tmp['port_type'] = val   # port_type
                    elif key == 'mode':
                        file_value = val.split('.')[0]
                        assert file_value == 'file' or file_value == 'directory',\
                            'port attribute value invalid.'
                        tmp['port_mode'] = val
                    elif key == 'multiplicity':
                        tmp['port_multiplicity'] = int(val)  # port_multiplicity
                    else:
                        assert False, 'invalid port attribute: %r=%r. fatal.'%\
                                (key,val)

                assert len(tmp) == 4
                store = (tmp['port_name'], tmp['port_type'], tmp['port_mode'],
                         tmp['port_multiplicity'])

                # (port_name, port_type, port_mode, port_multiplicity)
                __ports.append(store)

                # clear
                tmp   = None
                store = None

            if tag == 'diagram':

                self.__port_diagram = text

            if tag == 'ascii_art':

                self.__ascii_art = text

        return

#======================= end class Droplet: ======================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

This Quantity class is to be used with other classes in plant-level process modules.

For unit testing do at the linux command prompt:
    python quantity.py

Sat Sep  5 12:51:34 EDT 2015
"""
# *******************************************************************************
import os
import sys
import numpy as npy
# *******************************************************************************

class Quantity():

    # todo: this probably should not have a "value" for the same reason as Specie.
    #      this needs some thinking.

    def __init__(self,
                 name='null-quantity',
                 formalName='null-quantity',
                 value=float(0.0),  # float, int, or npy.ndarray (float or int)
                 unit='null-unit'
                 ):

        assert isinstance(name, str), 'oops not string.'
        self._name = name

        assert isinstance(formalName, str), 'oops not string.'
        self._formalName = formalName

        assert isinstance(value, float) or isinstance(value, int) or  \
               isinstance(value, npy.ndarray),'illegal value.'

        if isinstance(value, npy.ndarray):
           assert value.dtype[:4] == 'float' or value.dtype[:2] == 'int'

        self._value = value

        assert isinstance(name, str), 'oops not string.'
        self._unit = unit

        self._name = name
        self._formalName = formalName
        self._value = value
        self._unit = unit

        return

# *******************************************************************************

# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def SetName(self, n):
        self._name = n

    def get_name(self):
        return self._name
    name = property(get_name, SetName, None, None)

    def SetValue(self, v):
        self._value = v

    def GetValue(self):
        return self._value
    value = property(GetValue, SetValue, None, None)

    def SetFormalName(self, fn):
        self._formalName = fn

    def GetFormalName(self):
        return self._formalName
    formalName = property(GetFormalName, SetFormalName, None, None)

    def SetUnit(self, f):
        self._unit = f

    def GetUnit(self):
        return self._unit
    unit = property(GetUnit, SetUnit, None, None)

# *******************************************************************************
# Internal helpers

# *******************************************************************************
# Printing of data members
    def __str__(self):
        s = '\n\t Quantity(): \n\t name=%s; formalName=%s; value=%s[%s]'
        return s % (self.name, self.formalName, self.value, self.unit)

    def __repr__(self):
        s = '\n\t Quantity(): \n\t name=%s; formalName=%s; value=%s[%s]'
        return s % (self.name, self.formalName, self.value, self.unit)


# *******************************************************************************
# Usage: -> python interface.py
if __name__ == "__main__":
    print('Unit testing for Quantity')
